# -*- coding: utf-8 -*-
"""
iMSminer: A Data Processing and Machine Learning Package for Imaging Mass Spectrometry 
@author: Yu Tin Lin (yutinlin@stanford.edu)
@author: Haohui Bao (susanab20020911@gmail.com)
@author: Troy R. Scoggins IV (t.scoggins@ufl.edu)
@author: Boone M. Prentice (booneprentice@ufl.chem.edu)
License: Apache-2.0
"""
import math
import os
import subprocess
import sys
import warnings
from itertools import combinations, permutations

import cv2
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import psutil
import seaborn
import seaborn as sns
import sklearn
import statsmodels.api as sm
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
from numba import jit
from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression
from sklearn.manifold import TSNE
from sklearn.metrics import (
    calinski_harabasz_score,
    davies_bouldin_score,
    silhouette_score,
)
from sklearn.preprocessing import PolynomialFeatures
from statannotations.Annotator import Annotator
from statsmodels.formula.api import ols

from .utils import (
    draw_ROI,
    get_MS1_hits,
    make_image_1c,
    make_image_1c_njit,
    plot_image_at_point,
    repel_labels,
    significance,
    str2bool,
)


class DataAnalysis:
    """
    Performs data analysis on preprocessed intensity matrix and coordinate arrays. Mass alignment function wasa refactored from the python module msalign (https://github.com/lukasz-migas/msalign)

    Attributes
    ----------
    data_dir : str, user input
        Path pointing to directory containing preprocessed data
    fig_ratio : str, user input
        Text:figure ratio for rendered figures from options `small`, `medium`, and `large`
    df_pixel_all : pd.DataFrame
        Dataframe of pixels by peaks with coordinates, ROIs, and replicates with columns mapped to m/z array
    mz : pd.Series
        Series of m/z values for peaks index mapped to columns of df_pixel_all
    ROI_info : pd.DataFrame
        Dataframe of ROI coordinates with ROI label and replicate number
    ROI_num : int, user input
        Number of ROIs in dataset
    ROIs : str, user input
        ROI annotations from left to right, top to bottom
    img_array_1c : np.ndarray
        Collection of single-channgled ion images with positions mapped to x,y coordinates
    ion_type : str, user input
        Ion types of MS1 hits
    mass_diff : float, user input
        Mass differences of ion types from monoisotopic neutural
    MS1_db_path : str, user input
        Local file path of MS1 database
    mz_col : int, user input
        Number denoting column in database corresponding to exact mass of monoisotopic neutral
    ms1_df : pd.DataFrame
        Dataframe containing a table of MS1 hits with chemical information
    analyte_class : list
        List of analyte classes contained in column class_col of ms1_df
    df_mean_all : pd.DataFrame
        Dataframe of mean intensities with groups ROIs and replicates
    """

    def __init__(
            self
    ):
        detect_gpu = subprocess.run(['nvidia-smi', '--query-gpu=name', '--format=csv'],
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if detect_gpu.returncode == 0:
            print("GPU Detected:\n", detect_gpu.stdout.strip())

            test_library = ['cupy', 'cudf', 'cuml']
            for lib in test_library:
                try:
                    __import__(lib)
                    print(
                        f"{lib.capitalize()} is installed and imported successfully!")
                    self.gpu = True
                except ImportError:
                    print(
                        f"{lib.capitalize()} is not installed or could not be imported.")
                    self.gpu = False
                    break
        else:
            print("No GPU detected or NVIDIA driver not installed.")
            self.gpu = False

        try:
            __import__("numba")
            print("Numba is installed and imported successfully!")
            self.jit = True
        except ImportError:
            print("Numba is not installed or could not be imported.")
            self.jit = False

        self.data_dir = input(
            "\n\n Enter directory path containing image datasets: \n\n")

        fig_ratio_chosen = False
        while not fig_ratio_chosen:
            fig_ratio = input(
                "\n\n Render figures with `small`, `medium`, or `large` ratio? \n\n")
            if fig_ratio == "small":
                self.fig_ratio = 10
                fig_ratio_chosen = True
            elif fig_ratio == "medium":
                self.fig_ratio = 12.5
                fig_ratio_chosen = True
            elif fig_ratio == "large":
                self.fig_ratio = 15
                fig_ratio_chosen = True
            else:
                fig_ratio_chosen = False
                print("Select one of the options.")

    def load_preprocessed_data(
            self
    ):
        """
        import preprocessed intensity matrix and coordinate arrays, perform ROI annotation and selection, and store information for further data analysis
        """

        datasets = os.listdir(self.data_dir)
        datasets = np.asarray(datasets)[(np.char.find(
            datasets, "coords") == -1) & (np.char.find(datasets, "csv") != -1)]

        self._df_pixel_all = pd.DataFrame()
        ROI_edge = pd.DataFrame()

        for rep, dataset in enumerate(datasets):
            print(f"Importing {dataset}.")
            df_build = pd.read_csv(f"{self.data_dir}/{dataset}")
            print(f"Finished importing {dataset}.")

            df_build.rename(columns={'Unnamed: 0': 'mz'}, inplace=True)
            df_build = df_build.T
            if not hasattr(self, "_mz"):
                self._mz = df_build.loc['mz']
                self.mz = self._mz.copy()
            df_build.drop('mz', inplace=True)

            df_coords = pd.read_csv(
                f"{self.data_dir}/{dataset[:-4]}_coords.csv")
            df_coords.rename(columns={'0': 'x', '1': 'y'}, inplace=True)
            df_coords['x'] = df_coords['x'] - np.min(df_coords['x']) + 1
            df_coords['y'] = df_coords['y'] - np.min(df_coords['y']) + 1

            df_build.reset_index(drop=True, inplace=True)
            df_coords.reset_index(drop=True, inplace=True)

            df_build = pd.concat([df_build, df_coords[['x', 'y']]], axis=1)

            img_array_1c = make_image_1c(data_2darray=pd.concat([pd.Series(np.sum(
                df_build.iloc[:, :-2], axis=1)), df_build.iloc[:, -2:]], axis=1).to_numpy())
            self.img_array_1c = img_array_1c
            self.ROI_num = int(
                input("\n\n Enter number of ROIs for analysis: \n\n"))
            ROIs = input(
                "\n\n Enter labels for ROIs, from left to right, top to bottom? Separate ROI names by one space: \n\n")
            ROIs = ROIs.split(" ")

            # ROI selection
            if not any(module in sys.modules for module in ["google.colab", "notebook"]):
                ROI_exit = False
                while not ROI_exit:
                    ROI_dim_array = np.empty((0, 4))
                    for ROI in ROIs:
                        ROI_dim = cv2.selectROI(
                            'ROI_selection', img_array_1c[0], showCrosshair=True)
                        ROI_dim = np.asarray(ROI_dim)
                        ROI_dim_array = np.append(
                            ROI_dim_array, ROI_dim.reshape((1, -1)), axis=0)
                    cv2.destroyAllWindows()
                    ROI_sele = input("\n\n Keep ROI selecction? (yes/no) \n\n")
                    if ROI_sele == "yes":
                        ROI_exit = True
                    else:
                        ROI_exit = False
            else:
                plt.close("all")
                fig, ax = plt.subplots(figsize=(5, 5))
                ax.imshow(img_array_1c[0])
                plt.show()

            df_build['ROI'] = 'placeholder'
            df_build['ROI_num'] = 'placeholder'
            df_build['replicate'] = rep

            for i, ROI in enumerate(ROIs):
                print(f"Select ROI {i}, labeled {ROI}.")
                try:
                    bottom = ROI_dim_array[i][1]
                    top = ROI_dim_array[i][1] + ROI_dim_array[i][3]
                    left = ROI_dim_array[i][0]
                    right = ROI_dim_array[i][0] + ROI_dim_array[i][2]

                    ROI_xy = (df_build['x'] >= left) & (df_build['x'] < right) & (
                        df_build['y'] >= bottom) & (df_build['y'] < top)
                    df_build.loc[ROI_xy, "ROI"] = ROI
                    df_build.loc[ROI_xy, "ROI_num"] = np.sum(
                        df_build["ROI"] == ROI)
                    ROI_edge = pd.concat([ROI_edge, pd.Series(
                        [ROI, bottom, top, left, right, rep])], axis=1)
                except:
                    # ROI selection for jupyter notebook and Google colab
                    left = int(
                        input("\n\n Enter lowest value on x (horizontal) coordinate: \n\n"))
                    right = int(
                        input("\n\n Enter highest value on x (horizontal) coordinate: \n\n"))
                    bottom = int(
                        input("\n\n Enter lowest value on y (vertical) coordinate: \n\n"))
                    top = int(
                        input("\n\n Enter highest value on y (vertical) coordinate: \n\n"))

                    ROI_xy = (df_build['x'] >= left) & (df_build['x'] < right) & (
                        df_build['y'] >= bottom) & (df_build['y'] < top)
                    df_build.loc[ROI_xy, "ROI"] = ROI
                    ROI_edge = pd.concat([ROI_edge, pd.Series(
                        [ROI, bottom, top, left, right, rep])], axis=1)

                print(f"ROI {i} is selected and labeled {ROI}!")
            self._df_pixel_all = pd.concat([self._df_pixel_all, df_build])

        self._df_pixel_all.reset_index(drop=True, inplace=True)
        self._df_pixel_all.columns = self._df_pixel_all.columns.astype(str)
        self.df_pixel_all = self._df_pixel_all.copy()
        ROI_edge = ROI_edge.T.reset_index(drop=True)
        ROI_edge.columns = ["ROI", "bottom",
                            "top", "left", "right", "replicate"]
        self.ROI_info = ROI_edge

    def calibrate_mz(
            self
    ):
        """
        Interactive calibration using polynomial regression via linear model of user-specified degree

        Parameters
        ----------
        degree : int, user input
            Degree for linear model used in polynomial regression calibration
        reference_mz : 1darray, user input
            Reference massess to perform calibration on
        calibration_exit : str, user input
            Exits calibration if user specifies `yes`
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/calibration"):
            os.makedirs(f"{self.data_dir}/figures/calibration")
        plt.style.use('default')

        exit_regression = False
        while not exit_regression:
            degree = int(
                input("\n\n Enter the degree of linear model for polynomial calibration. Enter an integer: \n\n"))
            reference_mz = input(
                "\n\n Enter reference m/z's for calibration. Separate each reference m/z by one space: \n\n")
            reference_mz = reference_mz.split(" ")
            reference_mz = np.asarray(reference_mz).astype(np.float32)
            resid_index = np.argmin(
                abs((self.mz.to_numpy()[:, np.newaxis] - reference_mz[np.newaxis, :])), axis=0)
            resid = self.mz[resid_index].to_numpy() - reference_mz

            poly = PolynomialFeatures(degree=degree, include_bias=False)
            poly_features_resid = poly.fit_transform(
                self.mz[resid_index].to_numpy().reshape(-1, 1))
            poly_reg_model = LinearRegression()
            poly_reg_model.fit(poly_features_resid, resid)
            R2 = poly_reg_model.score(poly_features_resid, resid)

            poly_features = poly.fit_transform(
                self.mz.to_numpy().reshape(-1, 1))
            resid_predicted = poly_reg_model.predict(poly_features)

            fig, ax = plt.subplots(
                figsize=(self.fig_ratio, self.fig_ratio*0.7))
            ax.scatter(self.mz[resid_index],
                       resid_predicted[resid_index]/self.mz[resid_index]*10**6, color="red")
            resid_index_compl = np.ones(self.mz.shape[0], dtype=bool)
            resid_index_compl[resid_index] = False
            ax.scatter(self.mz[resid_index_compl],
                       np.zeros_like(self.mz[resid_index_compl]), color="blue", alpha=0.25)
            ax.plot(self.mz[resid_index], resid_predicted[resid_index] /
                    self.mz[resid_index]*10**6, color='black')
            ax.set_title(f'Linear model of degree {degree}', weight="bold")
            ax.set_xlabel('Experimental m/z', weight="bold")
            ax.set_ylabel('Residual [ppm]', weight="bold")
            ax.text(0.5, 0.95, f'$R^2 = {R2:.3f}$', fontsize=12, horizontalalignment='center',
                    verticalalignment='top', transform=plt.gca().transAxes)
            plt.show()

            calibration_exit = input(
                "\n\n Accept changes to calibration? (yes/no/exit) \n\n")
            if calibration_exit == "yes":
                self.mz -= resid_predicted
                fig.savefig(
                    f"{self.data_dir}/figures/calibration/accepted_calibration_degree{degree}.png")
                exit_regression = True
            elif calibration_exit == "no":
                exit_regression = False
                plt.close(fig)
            elif calibration_exit == "exit":
                plt.close(fig)
                break

    def MS1_search(
            self,
            ppm_threshold: float = 5,
            MS1_search_method: str = "avg_spectrum",
            filter_db: bool = True,
            percent_RAM: float = 5
    ):
        """MS1 accurate mass search using database from user

        Parameters
        ----------
        ppm_threshold : int, optional
            ppm threshold for MS1 accurate mass hits, by default 5
        MS1_search_method : str, optional
            Method for MS1 hits, by default `avg_spectrum`
            Method `avg_spectrum` performs MS1 search against an average spectrum
            Method `multi_spectrum` performs MS1 search against a collection of spectra stored in a csv file
        filter_db : bool, optional
            Filters MS1 database prior to accurate mass search if `filter_db = True`, by default True
        percent_RAM : float, optional
            Percent available RAM to define size of chunking, by deafult 5
        """
        ion_type = input(
            "\n\n Enter ion types of interest. Separate each ion type by one space: \n\n")
        self.ion_type = ion_type.split(" ")
        mass_diff = input(
            "\n\n Enter the corresponding mass difference of ion type(s) of interest. List ion types in same order as previously and separate each ion type by one space: \n\n")
        mass_diff = mass_diff.split(" ")
        self.mass_diff = np.asarray(mass_diff).astype(np.float32)

        MS1_db_path = input(
            "\n\n Enter file path of MS1 database in csv format: \n\n")
        mz_col = int(input(
            "\n\n Enter column in MS1 database that corresponds to neurtral monoisotopic masses: \n\n")) - 1
        print("Importing MS1 database. . .")
        MS1_db = pd.read_csv(MS1_db_path)
        print("Finished importing MS1 database!")
        print(f"The columns of imported database are {MS1_db.columns}")

        filter_db = str2bool(
            input("\n\n Filter database? Recommended for speed. Enter (yes / no) \n\n"))

        # reduce database
        if filter_db:
            self.col_filter = int(
                input("\n\n Which column [number] in database to perform filtering? \n\n")) - 1
            filter_by = input(
                f"\n\n Enter the groups of interest for filtering column {self.col_filter+1} in MS1 database. Separate each group by one space: \n\n")
            self.filter_by = filter_by.split(" ")
            MS1_db = MS1_db.loc[MS1_db.iloc[:, self.col_filter].str.contains(
                '|'.join(self.filter_by))]
            MS1_db.reset_index(drop=True, inplace=True)

        all_rows = []

        MS1_db.rename(
            columns={MS1_db.columns[mz_col]: 'exact_mass'}, inplace=True)

        if MS1_search_method == "avg_spectrum":
            mz = self.mz.copy()
            mz = np.asarray(mz).reshape([-1, 1]).astype(np.float32)
            for mz_diff, adduct in zip(self.mass_diff, self.ion_type):
                MS1 = MS1_db.copy()
                MS1["ion_mass"] = MS1["exact_mass"] + mz_diff
                MS1["ion_type"] = adduct
                MS1_db_array = np.asarray(MS1)
                MS1_ion_mass = MS1["ion_mass"].to_numpy().astype(np.float32)
                MS1_hit_indices, ppm = get_MS1_hits(
                    mz, MS1_ion_mass, ppm_threshold)
                for analyte_pointer, ion_index in zip(*MS1_hit_indices):
                    combined_row = np.append(
                        [int(analyte_pointer), ppm[analyte_pointer, ion_index]], MS1_db_array[ion_index])
                    all_rows.append(combined_row)
        elif MS1_search_method == "multi_spectrum":
            mz_list_path = input(
                "\n\n Enter the file path of list of m/z's: \n\n")
            mz = pd.read_csv(mz_list_path)
            analyte_index = np.repeat(np.arange(mz.shape[0]), mz.shape[1]).reshape(
                [-1, 1]).astype(np.float32)
            mz = mz.to_numpy().reshape([-1, 1]).astype(np.float32)
            mz = np.column_stack((analyte_index, mz))
            MS1_db_array = np.asarray(MS1_db)
            RAM_available = psutil.virtual_memory()[1]
            num_chunks = int(mz.shape[0] * MS1_db_array.shape[0] *
                             np.dtype(np.float32).itemsize / (RAM_available * percent_RAM/100)) + 1
            chunk_size_base = mz.shape[0] // num_chunks
            remainder = mz.shape[0] % num_chunks
            for mz_diff, adduct in zip(self.mass_diff, self.ion_type):
                MS1 = MS1_db.copy()
                MS1["ion_mass"] = MS1["exact_mass"] + mz_diff
                MS1["ion_type"] = adduct
                MS1_db_array = np.asarray(MS1)
                MS1_ion_mass = MS1["ion_mass"].to_numpy().astype(np.float32)
                chunk_start = 0
                for i in range(num_chunks):
                    chunk_size_temp = chunk_size_base
                    if remainder > i:
                        chunk_size_temp += 1
                    print(f'mz_chunk: {i}; chunck_size: {chunk_size_temp}')
                    if i != 0:
                        chunk_start += chunk_size_temp
                    mz_chunk = mz[chunk_start:(chunk_start+chunk_size_temp)]
                    MS1_hit_indices, ppm = get_MS1_hits(
                        mz_chunk[:, 1].reshape([-1, 1]), MS1_ion_mass, ppm_threshold)
                    for analyte_pointer, ion_index in zip(*MS1_hit_indices):
                        combined_row = np.append(
                            [int(mz[analyte_pointer+chunk_start, 0]), ppm[analyte_pointer, ion_index]], MS1_db_array[ion_index])
                        all_rows.append(combined_row)

        self.ms1_df = pd.DataFrame(
            all_rows, columns=['analyte', 'ppm'] + MS1.columns.to_list())

        if MS1_search_method == "multi_spectrum":
            self.ms1_df = self.ms1_df.groupby(list(self.ms1_df.columns[[0, 2, 3, 4, 5, 6, 7, 11, 12]])).agg(
                frequency=('ppm', 'size'),
                ppm_mean=('ppm', 'mean'),
                ppm_std=('ppm', 'std')
            ).reset_index()
            self.ms1_df['frequency'] *= (np.max(mz[:, 0])+1)/mz.shape[0]*100

    def filter_analytes(
            self,
            method: str = "MS1"
    ):
        """
        Subset peak-picked untargeted data to analytes of interest

        Parameters
        ----------
        method : str, optional
            Type of filtering, by default "MS1"
            Method `MS1" subsets` untargeted data to MS1 hits
            Method `analyte_class` subsets untargeted data to analyte classes from MS1 hits
        """
        print(f"Filtering analytes by {method}.")
        truncate_col = np.where(self.df_pixel_all.columns == "x")[0][0]
        if method == "MS1":
            pass
        elif method == "MS2":
            MS2_path = input("\n\n Enter file path of MS2 results: \n\n")
            MS2_results = pd.read_csv(MS2_path)

            if not any(self.ms1_df.columns == "adduct"):
                ion_type = np.unique(self.ms1_df['ion_type'])
                adduct_mapping = input(
                    f"\n\n Map {ion_type} to the following. Separate each value by one space: \n\n")
                adduct_mapping = adduct_mapping.split(" ")
                mapping_dict = dict(zip(ion_type, adduct_mapping))
                self.ms1_df['adduct'] = self.ms1_df['ion_type'].replace(
                    mapping_dict).fillna('unknown')
            analyte_col = int(input(f"\n\n Enter column number that corresponds to analyte IDs. The columns are \n"
                              f"{self.ms1_df.columns}: \n\n")) - 1

            self.ms1_df = pd.merge(self.ms1_df, MS2_results, left_on=[f'{self.ms1_df.columns[analyte_col]}', 'adduct'], right_on=[
                                   'bulk_ID', 'adduct'], how='inner', suffixes=('', '_ms2'))
            self.ms1_df = self.ms1_df.iloc[self.ms1_df[[
                "adduct", "ns_ID"]].drop_duplicates().index].reset_index(drop=True)
        elif method == "analyte_class":
            print(f"The columns in MS1 dataframe are {self.ms1_df.columns}")
            class_col = int(input(
                f"\n\n Enter column [number] in MS1 dataframe corresponds to analyte class? The columns are \n"
                f"{self.ms1_df.columns}: \n\n")) - 1
            analyte_class = input(
                "\n\n Enter analyte classes of interest. Separate each analyte class by one space: \n\n")
            self.analyte_class = analyte_class.split(" ")
            self.ms1_df.loc[self.ms1_df.iloc[:, class_col].str.contains(
                '|'.join(analyte_class))].reset_index(drop=True, inplace=True)
        elif method == "ion_type":
            print(f"The columns in MS1 dataframe are {self.ms1_df.columns}")
            ion_col = int(
                input(f"\n\n Enter column [number] MS1 dataframe that corresponds to ion type. The columns are \n"
                      f"{self.ms1_df.columns}: \n\n")) - 1
            ion_type = input(
                "\n\n Enter ion types of interest. Separate each ion type by one space: \n\n")
            self.ms1_df.loc[self.ms1_df.iloc[:, ion_col].str.contains(
                '|'.join(ion_type))].reset_index(drop=True, inplace=True)
        else:
            print(
                "Method not recognized. Please choose from options specified in documentation. ")
            return print("No filtering was performed.")

        self.df_pixel_all = pd.concat([self.df_pixel_all.iloc[:, np.unique(
            self.ms1_df['analyte'])], self.df_pixel_all.iloc[:, truncate_col:]], axis=1)
        self.mz = self.mz[np.unique(self.ms1_df['analyte'])]
        self.mz.reset_index(drop=True, inplace=True)
        analyte_mapping = {old_analyte: new_analyte for new_analyte,
                           old_analyte in enumerate(np.unique(self.ms1_df['analyte']))}
        self.ms1_df['analyte'] = self.ms1_df['analyte'].squeeze().map(
            analyte_mapping)
        truncate_col = np.where(self.df_pixel_all.columns == "x")[0][0]
        self.df_pixel_all.columns = np.append(np.arange(truncate_col).astype(
            str), self.df_pixel_all.columns[truncate_col:])

    def normalize_pixel(
            self,
            method: str = "TIC"
    ):
        """Normalize pixels using specified method

        Parameters
        ----------
        method : str, optional
            Method to compute normalization factor, by default "TIC"
            Method `TIC` normalizes on total ion count over each pixel
            Method `RMS` normalizes on root mean square over each pixel
            Method `reference` normalizes on a reference analyte specified by m/z or index
            Method `max` normalizes on maximum intensity of each pixel
        """
        print(f"Normalizing pixels with {method}.")
        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
        if method == "TIC":
            self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].div(
                np.sum(self.df_pixel_all.iloc[:, :truncate_col], axis=1), axis=0
            )
        elif method == "RMS":
            self.df_pixel_all.iloc[:, :truncate_col] = np.sqrt(
                self.df_pixel_all.iloc[:, :truncate_col]**2 / self.df_pixel_all.shape[1])
        elif method == "reference":
            print(
                "Pixels with reference intensity = 0 are normalized on the reference mean")
            mz_reference = float(input("\n\n Enter m/z of reference mass: \n\n"))
            ref_col = np.argmin(np.abs(self.mz - mz_reference))
            self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].div(
                self.df_pixel_all.iloc[:, ref_col].apply(lambda x: self.df_pixel_all.iloc[:, ref_col].mean() if x == 0 else x), axis=0
            )
        elif method == "max":
            self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].div(
                np.max(self.df_pixel_all.iloc[:, :truncate_col], axis=1), axis=0
            )
        else:
            print("Normalization method not found. Default to no normalization")
            pass

    def get_ion_image(
            self,
            replicate: int = 0,
            show_ROI: bool = True,
            show_square: bool = True,
            color_scheme: str = "inferno",
            ROI_size_divisor: float = 8,
            quantile: float = 100
    ):
        """
        Render ion image for analytes in self._df_pixel_all (filtered or unfiltered)

        Parameters
        ----------
        replicate : int, optional
            Render image from replicate (dataset) #, by default 0
        show_ROI : bool, optional
            Display ROI label above redenred ion image, by default True
        show_square : bool, optional
            Display a green box around selected ROI in rendered ion image, by default False
        color_scheme : str, optional
            False-color scheme for ion image visualization, by default "inferno". A list of color schemes are available here: https://matplotlib.org/stable/users/explain/colors/colormaps.html
        ROI_size_divisor : float, optional
            Controls size of ROI labels, where a smaller divisor gives a larger ROI label, by default 8
        quantile : float, optional
            Quantile of intensity (possible values in [0, 100]) for ion image visualization, by default 100
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/ion_image"):
            os.makedirs(f"{self.data_dir}/figures/ion_image")
        resolution = float(
            input("\n\n Enter spatial resolution of imaging [microns]: \n\n"))

        plt.style.use('default')

        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0] + 2
        self.df_pixel_all.iloc[:, :truncate_col -
                               2] = self.df_pixel_all.iloc[:, :truncate_col-2].fillna(0)
        insitu_df = pd.concat([self.df_pixel_all['replicate'],
                               self.df_pixel_all.iloc[:, :truncate_col]],
                              axis=1)
        insitu_df = insitu_df[insitu_df['replicate'] == replicate]
        insitu_df = insitu_df.drop(columns=['replicate'])

        if self.jit:
            insitu_2darray = insitu_df.to_numpy()
            img_array_1c = np.zeros([insitu_2darray.shape[1]-2, int(max(insitu_2darray[:, -1]))+1,
                                     int(max(insitu_2darray[:, -2]))+1, 1])
            insitu_image = make_image_1c_njit(
                data_2darray=insitu_2darray, img_array_1c=img_array_1c, remap_coord=False)
            del insitu_2darray, img_array_1c
        else:
            insitu_image = make_image_1c(
                data_2darray=insitu_df.to_numpy(), remap_coord=False, max_normalize=False)

        ROI_info = self.ROI_info[self.ROI_info["replicate"] == replicate]
        ROI_info.reset_index(drop=True, inplace=True)
        show_ROI = True
        show_square = True
        width_ratios = (ROI_info['right']-ROI_info['left']
                        ).to_numpy().astype(np.float32)
        width_ratios /= max(width_ratios)
        width_ratios = width_ratios.tolist()
        plt.rcParams.update({'font.size': np.max(
            self.df_pixel_all[['x', 'y']]).max(axis=0)/20})
        ROI_label_size = max(
            ROI_info['right']-ROI_info['left']) / ROI_size_divisor
        for analyte, im in enumerate(insitu_image[:, :, :, 0]):
            plt.style.use('dark_background')
            fig = plt.figure()
            gs = GridSpec(1, len(
                ROI_info) + 1, width_ratios=width_ratios + [0.1], figure=fig, wspace=0.25)
            q_max = np.percentile(im, quantile)
            norm = Normalize(vmin=im.min(), vmax=q_max)
            im = norm(im)
            im = np.clip(im, 0, 1)
            ROI_max = 0
            for ax_index in range(len(ROI_info)):
                ax = fig.add_subplot(gs[0, ax_index])
                ROI = pd.DataFrame(ROI_info.loc[ax_index]).T
                start_col = np.where(ROI.columns == "bottom")[0][0]
                truncate_col = np.where(ROI_info.columns == "right")[0][0]
                squares = ROI.iloc[:, start_col:truncate_col+1].to_numpy()
                im_ROI = im[int(squares[0][0]):(int(squares[0][1])+1),
                            int(squares[0][2]):(int(squares[0][3])+1)]
                if np.max(im_ROI) > ROI_max:
                    ROI_max = np.max(im_ROI)
                ax.imshow(im_ROI, cmap=color_scheme, vmin=0, vmax=1)
                plt.sca(ax)
                for i, square in enumerate(squares):
                    bottom, top, left, right = square
                    top -= bottom
                    bottom -= bottom
                    right -= left
                    left -= left
                    x_coords = [left, left, right, right, left]
                    y_coords = [bottom, top, top, bottom, bottom]
                    if show_square:
                        plt.plot(x_coords, y_coords, 'g-', linewidth=3)
                    if show_ROI:
                        plt.text(right/2, bottom-1, ROI.iloc[0, 0],
                                 horizontalalignment='center', size=ROI_label_size, color='white')
                ax.set_xticks([])
                ax.set_yticks([])
                ax.axis('off')
            cax = fig.add_subplot(gs[0, -1])
            fig.suptitle(f"m/z {self.mz[analyte]:.3f}",
                         fontsize=ROI_label_size, fontweight='bold')
            scalar_mappable = ScalarMappable(norm=norm, cmap=color_scheme)
            scalar_mappable.set_clim(0, quantile/100)
            colorbar = plt.colorbar(
                scalar_mappable, cax=cax, orientation='vertical')
            colorbar.set_ticks([0, quantile/100])
            colorbar.set_ticklabels([0, quantile])
            cax.set_position([cax.get_position().x0, cax.get_position().y0 + cax.get_position(
            ).height * 0.25, cax.get_position().width, cax.get_position().height * 0.5])
            cax_pos = cax.get_position()
            line_y = cax_pos.y1 + 0.05
            line_x_start = cax_pos.x0
            line_x_end = cax_pos.x1
            line = Line2D([line_x_start, line_x_end], [line_y, line_y], color='white',
                          transform=fig.transFigure, clip_on=False, linewidth=ROI_label_size/10)
            fig.add_artist(line)
            resolution_plt = resolution * \
                max(round(
                    np.max(ROI_info['right']-ROI_info['left'])*(line_x_end-line_x_start), 0)*self.ROI_num, 1)
            fig.text((line_x_start + line_x_end) / 2, line_y + 0.02, f'{resolution_plt: .0f} $\mu$m',
                     ha='center', va='center', color='white', fontsize=ROI_label_size/2)
            plt.savefig(f"{self.data_dir}/figures/ion_image/mz_{self.mz[analyte]: .3f}_replicate{replicate}.png",
                        bbox_inches='tight')
            plt.show()

    def make_FC_plot(
            self,
            pthreshold: float = 0.05,
            FCthreshold: float = 1.5,
            legend_label: str = "condition",
            feature_label: str = "mz",
            hm_label: str = "mz",
            jitter_amount: float = 2,
            jitter_factor: float = 100,
            font_size: float = 15,
            get_hm: bool = True,
            hm_width_factor: float = 10,
            hm_height_factor: float = 30,
            hm_fontsize: float = 10,
            hm_wspace: float = 1.5
    ):
        """
        Generate volcano plots of permuted ROI pairs, showing fold-change statistics and p-values

        Parameters
        ----------
        pthreshold : float, optional
            P-value threshold for statistical significance in volcano plot, by default 0.05
        FCthreshold : float, optional
            Absolute fold change threshold for significant dysregulation in volcano plot, by default 1.5
        legend_label : str, optional
            Labeling scheme for legend of volcano plot, by default `condition`
            Method `condition` colors data points by expression condition
            Method `analyte_class` colors significant data points by class of analyte from MS1 search. Prequisite: DataAnalysis.MS1_search()
        feature_label : str, optional
            Labeling scehem for data points in volcano plot, by default `mz`
            Method `mz` labels significant data points by their corresponding m/z values
            Method `analyte` labels significant data points by their corresponding analylte IDs.Prequisite: DataAnalysis.MS1_search()
        hm_label : str, optional
            Labeling scheme for data points in heatmap, by default `mz`
            Method `mz` visualizes m/z values
            Method `analyte` visualizes analyte IDs. Prequisite: DataAnalysis.MS1_search()
        jitter_amount : float, optional
            Controls placement of feature label in volcano plot, by default 2
        jitter_factor : float, optional
            Controls the repulsiveness of neighboring feature labels, by deafult 100
        font_size : float, optional
            Font size of feature labels in volcano plot, by default 15
        get_hm : bool, optional
            Renders a heatmap of feature label by ROI with entries fold change if `get_hm = True`, by deafult True
        hm_width_factor : float, optional
            Controls width of heatmaps
        hm_height_factor : float, optional
            Controls height of heatmap
        hm_fontsize : float, optional
            Controls font size in heatmap
        hm_wspace : float ,optional
            Controls column spacing between heatmaps
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/volcano"):
            os.makedirs(f"{self.data_dir}/figures/volcano")
        if not os.path.exists(f"{self.data_dir}/figures/volcano_unlabeled"):
            os.makedirs(f"{self.data_dir}/figures/volcano_unlabeled")

        plt.style.use('default')
        if feature_label == "analyte":
            analyte_col = int(input(f"\n\n Enter column number that corresponds to analyte IDs. \n"
                                    f"The columns are {self.ms1_df.columns}: \n\n")) - 1
        if legend_label == "analyte_class":
            self.vp_class_colnum = int(input(
                f"\n\n Enter column [number] containing classes of analytes to label in legend of volcano plot. \n\n The columns are {self.ms1_df.columns}: \n\n")) - 1
            self.analyte_class_VP = input(
                "\n\n Enter classes of analytes to label in legend of volcano plot. Separate each class by one space: \n\n")
            self.analyte_class_VP = self.analyte_class_VP.split(" ")

        # prepare dataframe for p-value and FC computations
        self.df_mean_all = self.df_pixel_all.groupby(
            ['ROI', 'replicate', 'ROI_num']).mean().reset_index()
        self.df_mean_all.columns = self.df_mean_all.columns.astype(str)
        self.df_mean_all = self.df_mean_all[self.df_mean_all['ROI']
                                            != 'placeholder']
        if self.df_mean_all['ROI'].shape[0] < 2:
            warnings.warn("Needs at least 2 ROIs.", RuntimeWarning)
            return None

        df_mean_all_long = pd.melt(self.df_mean_all, id_vars=[
            'ROI', 'replicate', 'ROI_num', 'x', 'y'], var_name='analyte', value_name='intensity')
        df_mean_all_long['analyte'] = df_mean_all_long['analyte'].astype(int)
        df_FC = df_mean_all_long.groupby(
            ['ROI', 'analyte']).mean().reset_index()
        df_hm = df_FC.copy()

        print("Starting to generate volcano plots.")

        for combo in list(permutations(df_FC['ROI'].unique(), 2)):
            df_FC_pair = df_FC[(df_FC['ROI'] == combo[0]) |
                               (df_FC['ROI'] == combo[1])]
            if any(df_FC_pair['intensity'] == 0):
                print(f"Analyte(s) \n{self.mz[(df_FC_pair['intensity'] == 0).reset_index(drop=True)]} \nhave a zero mean intensity in \n"
                      f"{combo[0]}. Excluded from volcano plot {combo[0]} / {combo[1]}. ")
                exclude_analyte = df_FC_pair.loc[df_FC_pair['intensity'] == 0, "analyte"].to_numpy(
                )
                df_FC_pair = df_FC_pair.loc[~df_FC_pair['analyte'].isin(
                    exclude_analyte)]
            else:
                exclude_analyte = None
            reference_intensity = df_FC_pair[df_FC_pair['ROI'] == combo[1]].groupby('analyte')[
                'intensity'].mean()
            df_FC_pair = df_FC_pair.merge(
                reference_intensity, on='analyte', suffixes=('', '_ref'))
            df_FC_pair['intensity'] /= df_FC_pair['intensity_ref']

            df_FC_pair['intensity'] = np.log2(df_FC_pair['intensity'])

            df_FC_pair['p'] = 1
            # compute p-value for each analyte
            for analyte_sele in df_FC_pair['analyte'].unique():
                df_analyte = df_mean_all_long[df_mean_all_long['analyte']
                                              == analyte_sele]
                lm_df = ols('intensity ~ C(ROI)',
                            data=df_analyte).fit()
                if lm_df.df_resid == 0:
                    warnings.warn(
                        "Insufficient sample size for computing p-values.", RuntimeWarning)
                    return None

                anova_df = sm.stats.anova_lm(lm_df, typ=1)

                df_FC_pair.loc[df_FC_pair['analyte'] ==
                               analyte_sele, 'p'] = anova_df['PR(>F)'][0]

            df_FC_pair['p'] = -np.log10(df_FC_pair['p'])
            df_FC_pair['legend'] = "placeholder"
            df_vp = df_FC_pair[df_FC_pair['ROI']
                               == combo[0]].set_index('analyte')
            if legend_label == "condition":
                palette = {
                    "placeholder": "grey",
                    "downregulated": "blue",
                    "upregulated": "red"
                }
                df_vp.loc[(df_vp['p'] > -np.log10(pthreshold)) & (
                    df_vp['intensity'] < -np.log2(FCthreshold)), 'legend'] = "downregulated"
                df_vp.loc[(df_vp['p'] > -np.log10(pthreshold)) & (
                    df_vp['intensity'] > np.log2(FCthreshold)), 'legend'] = "upregulated"
            elif legend_label == "analyte_class":
                color_keys = list(mcolors.TABLEAU_COLORS.keys())[
                    :len(self.filter_by)]
                palette = dict(zip(self.filter_by, color_keys))
                palette = {"placeholder": "grey"} | palette
                vp_features = self.ms1_df.iloc[:, [
                    0, self.col_filter+1]].drop_duplicates().to_numpy()
                df_vp = df_vp.iloc[vp_features[:, 0].astype(int)]
                vp_indices = (df_vp['p'] > -np.log10(pthreshold)
                              ) & (abs(df_vp['intensity']) > np.log2(FCthreshold))
                df_vp.loc[vp_indices,
                          "legend"] = vp_features[vp_indices, 1]
            plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
            fig = plt.figure(figsize=(self.fig_ratio *
                                      1.5, self.fig_ratio*1.5))
            ax = fig.add_subplot()
            scatter_plt = sns.scatterplot(data=df_vp,
                                          x="intensity", y="p", hue="legend",
                                          palette=palette, s=80, ax=ax)
            colors = scatter_plt.collections[0].get_facecolors()
            ax.set_title(f'{combo[0]} / {combo[1]}', weight="bold")
            ax.set_xlabel('Log2 FC', weight="bold")
            ax.set_ylabel('-Log10 p_value', weight="bold")
            plt.axvline(x=-np.log2(FCthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            plt.axvline(x=np.log2(FCthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            plt.axhline(y=-np.log10(pthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            if feature_label == "mz":
                vp_indices = (df_vp['p'] > -np.log10(pthreshold)
                              ) & (abs(df_vp['intensity']) > np.log2(FCthreshold))
                try:
                    repel_labels(ax, df_vp['intensity'][vp_indices],  df_vp['p'][vp_indices],
                                 np.round(self.mz.iloc[np.unique(
                                     df_FC_pair['analyte'])].to_numpy(), 1)[vp_indices],
                                 colors[vp_indices], k=jitter_amount, jitter_factor=jitter_factor, font_size=font_size)
                except ValueError:
                    pass
            elif feature_label == "analyte":
                vp_retain = self.ms1_df.loc[self.ms1_df['analyte'] != exclude_analyte, 'analyte'].to_numpy().astype(
                    int)
                _, vp_retain = np.unique(vp_retain, return_inverse=True)
                df_vp_mapping = df_vp.iloc[vp_retain]
                colors = [color for color in colors[vp_retain]]
                vp_indices = (df_vp_mapping['p'] > -np.log10(pthreshold)) & (
                    abs(df_vp_mapping['intensity']) > np.log2(FCthreshold))
                try:
                    repel_labels(ax, df_vp_mapping['intensity'][vp_indices],  df_vp_mapping['p'][vp_indices],
                                 self.ms1_df.loc[self.ms1_df['analyte'] != exclude_analyte, self.ms1_df.columns[analyte_col]][vp_indices.reset_index(
                                     drop=True)], pd.Series(colors)[vp_indices.reset_index(drop=True)], k=jitter_amount,
                                 jitter_factor=jitter_factor, font_size=font_size)
                except ValueError:
                    pass
            elif feature_label == "none":
                pass
            handles, labels = scatter_plt.get_legend_handles_labels()
            filtered_handles = [h for h, l in zip(
                handles, labels) if l != 'placeholder']
            filtered_labels = [l for l in labels if l != 'placeholder']
            ax.legend(filtered_handles, filtered_labels,
                      title_fontproperties={'weight': 'bold'})

            plt.tight_layout()
            plt.savefig(
                f"{self.data_dir}/figures/volcano/{combo[0]}_{combo[1]}", dpi=300)
            plt.show()

            # vocalno unlabeled
            plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
            fig = plt.figure(
                figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
            ax = fig.add_subplot()
            scatter_plt = sns.scatterplot(data=df_vp,
                                          x="intensity", y="p", hue="legend",
                                          palette=palette, s=80, ax=ax)
            ax.set_title(f'{combo[0]} / {combo[1]}', weight="bold")
            ax.set_xlabel('Log2 FC', weight="bold")
            ax.set_ylabel('-Log10 p_value', weight="bold")
            plt.axvline(x=-np.log2(FCthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            plt.axvline(x=np.log2(FCthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            plt.axhline(y=-np.log10(pthreshold), color='black',
                        linewidth=1.5, linestyle='--')
            handles, labels = scatter_plt.get_legend_handles_labels()
            filtered_handles = [h for h, l in zip(
                handles, labels) if l != 'placeholder']
            filtered_labels = [l for l in labels if l != 'placeholder']
            ax.legend(filtered_handles, filtered_labels,
                      title_fontproperties={'weight': 'bold'})
            plt.tight_layout()
            plt.savefig(
                f"{self.data_dir}/figures/volcano_unlabeled/{combo[0]}_{combo[1]}", dpi=300)
            plt.show()

        if get_hm:
            df_hm = df_hm.pivot(
                index="analyte", columns="ROI", values="intensity")
            if hm_label == "mz":
                df_hm.set_index(self.mz[df_hm.index.to_numpy()], inplace=True)
            elif hm_label == "analyte":
                ID_col = int(input(
                    f"\n\n Enter the column [number] to use for labeling analytes in heatmap. The columns are {self.ms1_df.columns}: \n\n")) - 1
                df_hm = df_hm.iloc[self.ms1_df['analyte'].to_numpy().astype(
                    int)]
                df_hm.set_index(self.ms1_df.iloc[:, ID_col], inplace=True)

            df_hm = np.log2(df_hm.div(df_hm.mean(axis=1), axis=0))

            fig = plt.figure(figsize=(
                hm_height_factor*df_hm.shape[1]/df_hm.shape[0]*hm_width_factor, hm_height_factor))
            plt.rcParams.update({'font.size': hm_fontsize})
            if hm_label == "mz":
                gs = GridSpec(1, 1 + 1, width_ratios=[1] + [0.1],
                              figure=fig, wspace=hm_wspace)
                ax = fig.add_subplot(gs[0, 0])
                hm_plt = ax.imshow(df_hm.to_numpy(),
                                   cmap='coolwarm', aspect='auto')
                hm_plt.set_clim(-3, 3)

                for i in range(df_hm.shape[0]):
                    for j in range(df_hm.shape[1]):
                        ax.text(j, i, f'{df_hm.iloc[i, j]:.1f}',
                                ha='center', va='center', color='black')
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.set_yticks(np.arange(df_hm.shape[0]))
                ax.set_yticklabels(df_hm.index)
                ax.set_xticks(np.arange(df_hm.shape[1]))
                ax.set_xticklabels(df_hm.columns, rotation=90)
            elif hm_label == "analyte":
                self.col_sep = int(input(
                    f"\n\n Enter the column [number] that contains the groups of interest. The columns are {self.ms1_df.columns}: \n\n")) - 1
                filter_by = input(
                    f"\n\n Enter the groups of interest from {self.col_sep+1}. Separate each group by one space: \n\n")
                self.filter_by = filter_by.split(" ")
                gs = GridSpec(1, len(self.filter_by) + 1, width_ratios=[1]*len(
                    self.filter_by) + [0.1], figure=fig, wspace=hm_wspace)
                for k, analyte_class in enumerate(self.filter_by):
                    ax = fig.add_subplot(gs[0, k])
                    df_hm_subset = df_hm[(
                        self.ms1_df.iloc[:, self.col_sep] == analyte_class).to_numpy()]
                    hm_plt = ax.imshow(df_hm_subset.to_numpy(),
                                       cmap='coolwarm', aspect='auto')
                    hm_plt.set_clim(-3, 3)

                    for i in range(df_hm_subset.shape[0]):
                        for j in range(df_hm_subset.shape[1]):
                            ax.text(j, i, f'{df_hm_subset.iloc[i, j]:.1f}',
                                    ha='center', va='center', color='black')
                    ax.set_xlabel('')
                    ax.set_ylabel('')
                    ax.set_yticks(np.arange(df_hm_subset.shape[0]))
                    ax.set_yticklabels(df_hm_subset.index)
                    ax.set_xticks(np.arange(df_hm_subset.shape[1]))
                    ax.set_xticklabels(df_hm_subset.columns, rotation=90)
            cax = fig.add_subplot(gs[0, -1])
            colorbar = plt.colorbar(hm_plt, cax=cax)
            colorbar.set_ticks(range(-3, 3+1))
            colorbar.ax.tick_params(labelsize=15)
            cax.set_position([cax.get_position().x0, cax.get_position().y0 + cax.get_position(
            ).height * 0.3, cax.get_position().width, cax.get_position().height * 0.45])
            plt.savefig(f"{self.data_dir}/figures/volcano/heatmap",
                        dpi=100, bbox_inches='tight')
            plt.show()

    def insitu_clustering(
            self,
            k: int = 10,
            perplexity: float = 15,
            replicate: int = 0,
            show_ROI: bool = True,
            show_square: bool = True,
            ROI_linewidth: float = 3,
            ROI_size_divisor: float = 8,
            insitu_tsne: bool = False
    ):
        """
        In situ segmentation via k-means clusterin. Groups pixels by similarity in molecular profiles. Outputs in situ visualization of ROIs colored by cluster labels. In situ segmentation with cluster and ROI annotations are visualized in lower-dimensional t-SNE embedding.

        Parameters
        ----------
        k : int, optional
            Number of k-means clusters. The default is 10.
        perplexity : float, optional
            t-SNE embedding parameter, which influences tightness of embedded neighbors. The default is 15.
        replicate : int, optional
            Dataset # of which ion images are rendered. The default is 0.
        show_ROI : bool, optional
            Display ROI label above redenred ion image if True. The default is True.
        show_square : bool, optional
            Display a green box around selected ROI in rendered ion image if True. The default is False.
        ROI_line_width : float, optional
            Controls width of green box surrounding ROIs
        ROI_size_divisor : float, optional
            Controls size of ROI labels, where a smaller divisor gives a larger ROI label, by default 8
        insitu_tsne : bool, optional
            Renders a RGB in situ representation of 3D t-SNE embedding if `insitu_tsne = True`, by default False
        """
        resolution = float(
            input("\n\n Enter spatial resolution of imaging [microns]: \n\n"))
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/insitu_cluster"):
            os.makedirs(f"{self.data_dir}/figures/insitu_cluster")
        plt.style.use('default')
        # KMeans
        self.k = k
        self.perplexity = perplexity
        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
        self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].fillna(
            0)
        df_pixel_all_max = self.df_pixel_all.copy()
        df_pixel_all_max = df_pixel_all_max.loc[df_pixel_all_max["ROI"]
                                                != "placeholder"]
        df_pixel_all_max.reset_index(drop=True, inplace=True)
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=0), axis=1
        )
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=1), axis=0
        )
        df_pixel_all_max = df_pixel_all_max.fillna(0)
        num_iters = 1000

        if self.gpu == False:
            print("Spatial segmentation on CPU. Computing time may be long.")
            from sklearn.cluster import KMeans
            from sklearn.manifold import TSNE

            # tSNE
            k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                             n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col])
            kmeans_labels = k_means.labels_ + 1
            my_random_state = 1
            tsne = TSNE(n_components=2, n_iter=num_iters,
                        perplexity=self.perplexity, random_state=my_random_state)
            tsne_embedding = tsne.fit_transform(
                df_pixel_all_max.iloc[:, :truncate_col])

        elif self.gpu == True:
            try:
                print("Spatial segmentation on GPU.")
                from cuml import using_output_type
                from cuml.cluster import KMeans
                from cuml.manifold import TSNE

                with using_output_type('numpy'):
                    k_means = KMeans(n_clusters=self.k, max_iter=num_iters).fit(
                        df_pixel_all_max.iloc[:, :truncate_col])
                    kmeans_labels = k_means.labels_ + 1

                    my_random_state = 1
                    tsne = TSNE(n_components=2, n_iter=num_iters,
                                perplexity=self.perplexity, random_state=my_random_state,
                                n_neighbors=150)
                    tsne_embedding = tsne.fit_transform(
                        df_pixel_all_max.iloc[:, :truncate_col])

            except:
                print("GPU error. Defaulting to CPU.")
                print("Spatial segmentation on CPU. Computing time may be long.")
                from sklearn.cluster import KMeans
                from sklearn.manifold import TSNE
                k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                 n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col])
                kmeans_labels = k_means.labels_ + 1

                my_random_state = 1
                tsne = TSNE(n_components=2, n_iter=num_iters,
                            perplexity=self.perplexity, random_state=my_random_state)
                tsne_embedding = tsne.fit_transform(
                    df_pixel_all_max.iloc[:, :truncate_col])

        df_pixel_all_max['insitu_cluster'] = pd.DataFrame(kmeans_labels)

        plt.figure(figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
        plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
        cmap = plt.cm.get_cmap('rainbow', len(np.unique(kmeans_labels)))
        for i, label in enumerate(np.unique(kmeans_labels)):
            plt.scatter(tsne_embedding[kmeans_labels == label, 0], tsne_embedding[kmeans_labels == label, 1],
                        c=cmap(i),
                        alpha=0.6, label=f'Cluster {label}')

        plt.title(
            "In situ segmentation in t-SNE", weight='bold')
        plt.xlabel("t-SNE 1", weight='bold')
        plt.ylabel("t-SNE 2", weight='bold')
        legend = plt.legend(
            title='Cluster', title_fontproperties={'weight': 'bold'})
        for legend_handle in legend.legendHandles:
            legend_handle.set_alpha(1)
        plt.tight_layout()
        plt.savefig(
            f"{self.data_dir}/figures/insitu_cluster/kmeans_tsne", dpi=100)

        categories = pd.Categorical(df_pixel_all_max['ROI'])
        unique_categories = categories.categories
        cmap = plt.cm.get_cmap('rainbow', len(unique_categories))

        plt.figure(figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
        plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})

        for i, category in enumerate(unique_categories):
            mask = df_pixel_all_max['ROI'] == category
            plt.scatter(tsne_embedding[mask, 0], tsne_embedding[mask, 1],
                        color=cmap(i), alpha=0.6, label=category)

        plt.title("In situ segmentation in t-SNE", weight='bold')
        plt.xlabel("t-SNE 1", weight='bold')
        plt.ylabel("t-SNE 2", weight='bold')

        legend = plt.legend(
            title='ROI', title_fontproperties={'weight': 'bold'})
        for legend_handle in legend.legendHandles:
            legend_handle.set_alpha(1)
        plt.tight_layout()
        plt.savefig(
            f"{self.data_dir}/figures/insitu_cluster/original_space_tsne", dpi=100)

        insitu_df = pd.concat([df_pixel_all_max['replicate'], pd.DataFrame(kmeans_labels, columns=["kmeans_label"]),
                               df_pixel_all_max[['x', 'y']]], axis=1)
        insitu_df = insitu_df[insitu_df['replicate'] == replicate]
        insitu_df = insitu_df.drop(columns=['replicate'])

        insitu_image = make_image_1c(
            data_2darray=insitu_df.to_numpy(), remap_coord=False, max_normalize=False)
        insitu_image = np.ma.masked_equal(insitu_image, 0)

        ROI_info = self.ROI_info[self.ROI_info["replicate"] == replicate]
        ROI_info.reset_index(drop=True, inplace=True)

        show_ROI = True
        show_square = True
        width_ratios = (ROI_info['right']-ROI_info['left']
                        ).to_numpy().astype(np.float32)
        width_ratios /= max(width_ratios)
        width_ratios = width_ratios.tolist()
        plt.style.use('dark_background')
        fig = plt.figure()
        plt.rcParams.update({'font.size': np.max(
            self.df_pixel_all[['x', 'y']]).max(axis=0)/20})
        ROI_label_size = max(
            ROI_info['right']-ROI_info['left']) / ROI_size_divisor
        gs = GridSpec(1, len(ROI_info) + 1,
                      width_ratios=width_ratios + [0.1], figure=fig, wspace=0.25)
        for ax_index in range(len(ROI_info)):
            ax = fig.add_subplot(gs[0, ax_index])
            ROI = pd.DataFrame(ROI_info.loc[ax_index]).T
            start_col = np.where(ROI.columns == "bottom")[0][0]
            truncate_col = np.where(ROI_info.columns == "right")[0][0]
            squares = ROI.iloc[:, start_col:truncate_col+1].to_numpy()

            im_ROI = insitu_image[0][int(squares[0][0]):(int(squares[0][1])+1),
                                     int(squares[0][2]):(int(squares[0][3])+1)]
            kmeans_insitu = ax.imshow(
                im_ROI, cmap="rainbow", vmin=1, vmax=np.max(kmeans_labels))
            plt.sca(ax)
            for i, square in enumerate(squares):
                bottom, top, left, right = square
                top -= bottom
                bottom -= bottom
                right -= left
                left -= left
                x_coords = [left, left, right, right, left]
                y_coords = [bottom, top, top, bottom, bottom]
                if show_square:
                    plt.plot(x_coords, y_coords, 'g-', linewidth=3)
                if show_ROI:
                    plt.text(right/2, bottom-1, ROI.iloc[0, 0],
                             horizontalalignment='center', size=ROI_label_size, color='white')
            ax.set_xticks([])
            ax.set_yticks([])
            ax.axis('off')
        plt.title("")
        cax = fig.add_subplot(gs[0, -1])
        colorbar = plt.colorbar(kmeans_insitu, cax=cax)
        min_val, max_val = 1, np.max(kmeans_labels)
        colorbar.set_ticks(range(int(min_val), int(max_val) + 1))
        cax.set_position([cax.get_position().x0, cax.get_position().y0 + cax.get_position(
        ).height * 0.25, cax.get_position().width, cax.get_position().height * 0.5])
        cax_pos = cax.get_position()
        line_y = cax_pos.y1 + 0.05
        line_x_start = cax_pos.x0
        line_x_end = cax_pos.x1
        plt.tight_layout()
        line = Line2D([line_x_start, line_x_end], [line_y, line_y], color='white',
                      transform=fig.transFigure, clip_on=False, linewidth=ROI_linewidth)
        fig.add_artist(line)
        resolution_plt = resolution * \
            max(round(
                np.max(ROI_info['right']-ROI_info['left'])*(line_x_end-line_x_start), 0)*self.ROI_num, 1)
        fig.text((line_x_start + line_x_end) / 2, line_y + 0.02,
                 f'{resolution_plt:.0f} $\mu$m', ha='center', va='center', color='white', fontsize=ROI_label_size/2)

        plt.savefig(
            f"{self.data_dir}/figures/insitu_cluster/kmeans_insitu_image", dpi=200)
        plt.show(fig)

    def optimize_insitu_clustering(
            self,
            k_max: int = 10
    ):
        """
        In situ segmentation via k-means clusterin. Groups pixels by similarity in molecular profiles. Outputs in situ visualization of ROIs colored by cluster labels. In situ segmentation with cluster and ROI annotations are visualized in lower-dimensional t-SNE embedding.

        Parameters
        ----------
        k_max : int, optional
            Maximum number of clusters for cluster validity evaluation
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/insitu_cluster"):
            os.makedirs(f"{self.data_dir}/figures/insitu_cluster")
        score_array = np.zeros((k_max-1, 4))

        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
        self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].fillna(
            0)
        df_pixel_all_max = self.df_pixel_all.copy()
        df_pixel_all_max = df_pixel_all_max.loc[df_pixel_all_max["ROI"]
                                                != "placeholder"]
        df_pixel_all_max.reset_index(drop=True, inplace=True)
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=0), axis=1
        )
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=1), axis=0
        )
        df_pixel_all_max = df_pixel_all_max.fillna(0)
        num_iters = 1000
        for k in range(2, k_max+1):
            self.k = k
            if self.gpu == False:
                print("In situ clustering on CPU. Computing time may be long.")
                from sklearn.cluster import KMeans
                k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                 n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col])
                kmeans_labels = k_means.labels_ + 1

            elif self.gpu == True:
                try:
                    print("Image clustering on GPU.")
                    from cuml import using_output_type
                    from cuml.cluster import KMeans
                    with using_output_type('numpy'):

                        k_means = KMeans(n_clusters=self.k, max_iter=num_iters).fit(
                            df_pixel_all_max.iloc[:, :truncate_col])
                        kmeans_labels = k_means.labels_ + 1

                except:
                    print("GPU error. Defaulting to CPU.")
                    print("Image clustering on CPU. Computing time may be long.")
                    from sklearn.cluster import KMeans
                    k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                     n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col])
                    kmeans_labels = k_means.labels_ + 1

                score_array[k-2, 0] = davies_bouldin_score(
                    df_pixel_all_max.iloc[:, :truncate_col], kmeans_labels)
                score_array[k-2, 1] = calinski_harabasz_score(
                    df_pixel_all_max.iloc[:, :truncate_col], kmeans_labels)
                score_array[k-2, 2] = silhouette_score(
                    df_pixel_all_max.iloc[:, :truncate_col], k_means.fit_predict(df_pixel_all_max.iloc[:, :truncate_col]))
                score_array[k-2, 3] = k_means.inertia_
        plt.style.use('default')
        categories = ["DB", "CH", "Silhouette", "Elbow"]
        plt.rcParams.update({'font.size': 22})
        fig, ax = plt.subplots(figsize=(10, 7))
        colors = ['blue', 'green', 'red', 'purple', "black"]
        normalized_scores = np.divide(score_array, np.max(score_array, axis=0))
        x_values = np.arange(2, k_max+1)
        for i, (category, color) in enumerate(zip(categories, colors)):
            y_values = normalized_scores[:, i]
            plt.scatter(x_values, y_values, color=color, label=category)
            plt.plot(x_values, y_values, color=color)
        plt.xlabel('Number of clusters', weight='bold')
        plt.ylabel('Max-Normalized Scores', weight='bold')
        plt.legend(title='Score', fontsize="small",
                   title_fontproperties={'weight': 'bold'})
        plt.savefig(
            f"{self.data_dir}/figures/insitu_cluster/insitu_clustering_optimization_plot.png", bbox_inches='tight')
        plt.show()

    def image_clustering(
            self,
            k: int = 10,
            perplexity: float = 5,
            replicate: int = 0,
            show_ROI: bool = True,
            show_square: bool = True,
            color_scheme: str = "inferno",
            insitu_tsne: bool = False,
            insitu_perplexity: float = 15,
            zoom: float = 0.1,
            quantile: float = 100,
            img_plot_method: str = "plot_ROI",
            feature_label="mz",
            jitter_amount: float = 2,
            jitter_factor: float = 100,
            font_size: float = 15,
            ROI_linewidth: float = 3,
            ROI_size_divisor: float = 8
    ):
        """
        Group ion images by spatial co-localization. Outputs in situ and box plot visualizations of mean ion image for each cluster. Clustering analysis and in situ mapping of clusters are summarized in lower-dimensional t -SNE embedding.

        Parameters
        ----------
        k : int, optional
            Number of k-means clusters. The default is 10.
        perplexity : float, optional
            t-SNE embedding parameter, which influences tightness of embedded neighbors. The default is 5.
        replicate : int, optional
            Dataset # of which ion images are rendered. The default is 0.
        show_ROI : bool, optional
            Display ROI label above redenred ion image if True. The default is True.
        show_square : bool, optional
            Display a green box around selected ROI in rendered ion image if True. The default is False.
        color_scheme : str, optional
            False-color scheme for ion image visualization. The default is "inferno". A list of color schemes are available here: https://matplotlib.org/stable/users/explain/colors/colormaps.html
        zoom : float, optional
            Relative size of ion images in t-SNE embedding to the embedding. The default is 0.1.
        quantile : TYPE, optional
            Maximum intensity quantile cutoff for ion image visualization. The default is 100.
        img_plot_method: str, optional
            Controls layout of rendered in situ heatmaps
            Method `plot_img` retains all coordinates from imaging mass spectrometry experiment
            Method `plot_ROI` renders heatmaps of ROIs 
        jitter_amount : float, optional
            Controls placement of feature labels in t-SNE embedding of ion images, by default 2
        jitter_factor : float, optional
            Controls the repulsiveness of neighboring feature labels in t-SNE embedding of ion images, by deafult 100
        font_size : float, optional
            Font size of feature labels in volcano plot, by default 15
        ROI_line_width : float, optional
            Controls width of green box surrounding ROIs, by default 3
        ROI_size_divisor : float, optional
            Controls size of ROI labels, where a smaller divisor gives a larger ROI label, by default 8
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/image_cluster"):
            os.makedirs(f"{self.data_dir}/figures/image_cluster")
        if feature_label == "analyte":
            analyte_col = int(input(f"\n\n Enter column number that corresponds to analyte IDs. The columns are \n"
                              f"{self.ms1_df.columns}: \n\n")) - 1

        plt.style.use('default')
        resolution = float(
            input("\n\n Enter spatial resolution of imaging [microns]: \n\n"))

        self.k = k
        self.perplexity = perplexity

        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
        self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].fillna(
            0)
        df_pixel_all_max = self.df_pixel_all.copy()
        df_pixel_all_max = df_pixel_all_max[df_pixel_all_max["ROI"]
                                            != "placeholder"]
        df_pixel_all_max.reset_index(drop=True, inplace=True)
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=0))
        num_iters = 1000
        if self.gpu == False:
            print("Image clustering on CPU. Computing time may be long.")
            from sklearn.cluster import KMeans
            from sklearn.manifold import TSNE
            k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                             n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col].T)
            kmeans_labels = k_means.labels_ + 1

            my_random_state = 1
            tsne = TSNE(n_components=2, n_iter=num_iters,
                        perplexity=self.perplexity, random_state=my_random_state)
            tsne_embedding = tsne.fit_transform(
                df_pixel_all_max.iloc[:, :truncate_col].T)
        elif self.gpu == True:

            try:
                print("Image clustering on GPU.")
                from cuml import using_output_type
                from cuml.cluster import KMeans
                from cuml.manifold import TSNE
                with using_output_type('numpy'):

                    k_means = KMeans(n_clusters=self.k, max_iter=num_iters).fit(
                        df_pixel_all_max.iloc[:, :truncate_col].T)
                    kmeans_labels = k_means.labels_ + 1

                    my_random_state = 1
                    tsne = TSNE(n_components=2, n_iter=num_iters,
                                perplexity=self.perplexity, random_state=my_random_state,
                                n_neighbors=150)
                    tsne_embedding = tsne.fit_transform(
                        df_pixel_all_max.iloc[:, :truncate_col].T)
            except:
                print("GPU error. Defaulting to CPU.")
                print("Image clustering on CPU. Computing time may be long.")
                from sklearn.cluster import KMeans
                from sklearn.manifold import TSNE
                k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                 n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col].T)
                kmeans_labels = k_means.labels_ + 1
                my_random_state = 1
                tsne = TSNE(n_components=2, n_iter=num_iters,
                            perplexity=self.perplexity, random_state=my_random_state)
                tsne_embedding = tsne.fit_transform(
                    df_pixel_all_max.iloc[:, :truncate_col].T)

        image_cluster_label = pd.DataFrame(kmeans_labels)
        image_cluster_label = image_cluster_label.reset_index()
        image_cluster_label.rename(
            columns={'index': 'analyte', 0: 'cluster'}, inplace=True)

        image_df = df_pixel_all_max.groupby(
            ['ROI', 'replicate', 'ROI_num']).mean().reset_index()
        cluster_mean_df = image_df['ROI']
        image_cluster_all = df_pixel_all_max['ROI']
        for cluster_index in range(1, k+1):  # loop through clusters
            cluster_df = image_df[image_cluster_label[image_cluster_label['cluster']
                                                      == cluster_index]['analyte'].to_numpy().astype(str)].mean(axis=1)
            cluster_df.name = str(cluster_index)
            cluster_mean_df = pd.concat([cluster_mean_df, cluster_df], axis=1)

            image_cluster = df_pixel_all_max[image_cluster_label[image_cluster_label['cluster']
                                                                 == cluster_index]['analyte'].to_numpy().astype(str)].mean(axis=1)
            image_cluster.name = str(cluster_index)
            image_cluster_all = pd.concat(
                [image_cluster_all, image_cluster], axis=1)

        cluster_mean_df = cluster_mean_df[cluster_mean_df['ROI']
                                          != "placeholder"]

        image_cluster_all[['x', 'y']] = df_pixel_all_max[['x', 'y']]
        image_cluster_all = image_cluster_all[df_pixel_all_max['replicate'] == replicate]

        plt.style.use('default')
        plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
        fig = plt.figure(figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
        ax = fig.add_subplot()
        cmap = plt.cm.get_cmap('rainbow', len(np.unique(kmeans_labels)))
        label_to_color = {label: cmap(
            i) for i, label in enumerate(np.unique(kmeans_labels))}
        for i, label in enumerate(np.unique(kmeans_labels)):
            ax.scatter(tsne_embedding[kmeans_labels == label, 0], tsne_embedding[kmeans_labels == label, 1],
                       c=cmap(i),
                       alpha=1, label=f'Cluster {label}', s=50)
        if feature_label == "mz":
            flat_colors = [label_to_color[label] for label in kmeans_labels]
            repel_labels(ax, tsne_embedding[:, 0], tsne_embedding[:, 1],
                         np.round(self.mz.to_numpy(), 1), flat_colors, k=jitter_amount, jitter_factor=jitter_factor, font_size=font_size)
        elif feature_label == "analyte":
            ms1_features = self.ms1_df.iloc[:, [
                0, analyte_col]].drop_duplicates().to_numpy()
            tsne_mapping = tsne_embedding[ms1_features[:, 0].astype(int),]
            flat_colors = [label_to_color[label]
                           for label in kmeans_labels[ms1_features[:, 0].astype(int)]]
            repel_labels(ax, tsne_mapping[:, 0], tsne_mapping[:, 1],
                         ms1_features[:, 1], flat_colors, k=jitter_amount,
                         jitter_factor=jitter_factor, font_size=font_size)
        elif feature_label == "none":
            pass
        else:
            print("Method for `feature_label` not found. Defaults to no labeling.")
            pass

        ax.set_title(
            "Image clusters in t-SNE", weight='bold')
        ax.set_xlabel("t-SNE 1", weight='bold')
        ax.set_ylabel("t-SNE 2", weight='bold')
        legend = plt.legend(
            title='Cluster', title_fontproperties={'weight': 'bold'})
        for legend_handle in legend.legendHandles:
            legend_handle.set_alpha(1)
        plt.tight_layout()
        plt.savefig(
            f"{self.data_dir}/figures/image_cluster/image_kmeans_tsne", dpi=100)
        plt.show()

        truncate_col = np.where(df_pixel_all_max.columns == 'x')[0][0] + 2
        insitu_df = pd.concat([df_pixel_all_max['replicate'],
                               df_pixel_all_max.iloc[:, :truncate_col]],
                              axis=1)
        insitu_df = insitu_df[insitu_df['replicate'] == replicate]
        insitu_df = insitu_df.drop(columns=['replicate'])

        if self.jit:
            insitu_2darray = insitu_df.to_numpy()
            img_array_1c = np.zeros([insitu_2darray.shape[1]-2, int(max(insitu_2darray[:, -1]))+1,
                                     int(max(insitu_2darray[:, -2]))+1, 1])
            insitu_image = make_image_1c_njit(
                data_2darray=insitu_2darray, img_array_1c=img_array_1c, remap_coord=False)
            del insitu_2darray, img_array_1c
        else:
            insitu_image = make_image_1c(
                data_2darray=insitu_df.to_numpy(), remap_coord=False, max_normalize=False)

        if img_plot_method == "plot_img":
            plt.figure(figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
            plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
            plt.scatter(*tsne_embedding.T)
            for im, xy in zip(insitu_image[:, :, :, 0], tsne_embedding):
                q_max = np.percentile(im, quantile)
                norm = Normalize(vmin=im.min(), vmax=q_max)
                im = norm(im)
                im = np.clip(im, 0, 1)
                plot_image_at_point(im, xy, zoom=zoom, color_scheme="inferno")
                plt.title("Image clusters in t-SNE", weight='bold')
                plt.xlabel("t-SNE 1", weight='bold')
                plt.ylabel("t-SNE 2", weight='bold')
            plt.tight_layout()
            plt.savefig(
                f"{self.data_dir}/figures/image_cluster/image_insitu_tsne")
        elif img_plot_method == "plot_ROI":
            img_list = []
            ROI_max = 0
            ROI_info = self.ROI_info[self.ROI_info["replicate"] == replicate]
            ROI_info.reset_index(drop=True, inplace=True)
            show_ROI = True
            show_square = True
            color_scheme = "inferno"
            width_ratios = (
                ROI_info['right']-ROI_info['left']).to_numpy().astype(np.float32)
            width_ratios /= max(width_ratios)
            width_ratios = width_ratios.tolist()
            plt.rcParams.update({'font.size': np.max(
                df_pixel_all_max[['x', 'y']]).max(axis=0)/20})
            ROI_label_size = max(
                ROI_info['right']-ROI_info['left']) / ROI_size_divisor
            truncate_col = np.where(ROI_info.columns == "right")[0][0]

            for analyte, im in enumerate(insitu_image[:, :, :, 0]):
                plt.style.use('dark_background')
                q_max = np.percentile(im, quantile)
                norm = Normalize(vmin=im.min(), vmax=q_max)
                im = norm(im)
                im = np.clip(im, 0, 1)
                fig = plt.figure()
                gs = GridSpec(
                    1, len(ROI_info), width_ratios=width_ratios, figure=fig, wspace=0.25)

                for ax_index in range(len(ROI_info)):
                    ax = fig.add_subplot(gs[0, ax_index])
                    ROI = pd.DataFrame(ROI_info.loc[ax_index]).T
                    start_col = np.where(ROI.columns == "bottom")[0][0]
                    squares = ROI.iloc[:, start_col:truncate_col+1].to_numpy()

                    im_ROI = im[int(squares[0][0]):(int(squares[0][1])+1),
                                int(squares[0][2]):(int(squares[0][3])+1)]
                    if np.max(im_ROI) > ROI_max:
                        ROI_max = np.max(im_ROI)
                    ax.imshow(im_ROI, cmap=color_scheme)
                    plt.sca(ax)
                    for i, square in enumerate(squares):
                        bottom, top, left, right = square
                        top -= bottom
                        bottom -= bottom
                        right -= left
                        left -= left
                        x_coords = [left, left, right, right, left]
                        y_coords = [bottom, top, top, bottom, bottom]
                        if show_square:
                            plt.plot(x_coords, y_coords, 'g-', linewidth=3)
                        if show_ROI:
                            plt.text(right/2, bottom-1, ROI.iloc[0, 0],
                                     horizontalalignment='center', size=ROI_label_size, color='white')
                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.axis('off')

                fig.canvas.draw()
                plt.close()
                img_array = np.frombuffer(
                    fig.canvas.buffer_rgba(), dtype=np.uint8)
                img_array = img_array.reshape(
                    fig.canvas.get_width_height()[::-1] + (4,))
                img_list.append(img_array)

            plt.style.use('default')
            plt.figure(figsize=(self.fig_ratio*1.5, self.fig_ratio*1.5))
            plt.rcParams.update({'font.size': 200/self.fig_ratio*1.5})
            plt.scatter(*tsne_embedding.T)
            for im, xy in zip(img_list, tsne_embedding):
                plot_image_at_point(im, xy, zoom=zoom, color_scheme="inferno")
                plt.title("Image clusters in t-SNE", weight='bold')
                plt.xlabel("t-SNE 1", weight='bold')
                plt.ylabel("t-SNE 2", weight='bold')
            plt.savefig(
                f"{self.data_dir}/figures/image_cluster/image_insitu_tsne", bbox_inches='tight')
            plt.show()

        plt.style.use('default')
        image_cluster_df_long = pd.melt(cluster_mean_df, id_vars=[
                                        'ROI'], var_name='cluster', value_name='intensity')
        image_cluster_df_long['intensity'] = image_cluster_df_long['intensity'].div(
            image_cluster_df_long.groupby(['cluster']).transform('max')['intensity'])

        if np.unique(image_cluster_df_long['ROI']).shape[0] >= 2:
            ROI = image_cluster_df_long["ROI"]
            for cluster in image_cluster_df_long['cluster'].unique():
                image_cluster_df_cluster = image_cluster_df_long[
                    image_cluster_df_long['cluster'] == cluster]
                pvalue = [
                    sm.stats.anova_lm(
                        ols('intensity ~ C(ROI)', data=image_cluster_df_cluster.loc[
                            (image_cluster_df_cluster['ROI'] == combo[0]) |
                            (image_cluster_df_cluster['ROI'] == combo[1]),
                            ['ROI', 'intensity']]
                            ).fit(),
                        typ=1
                    )['PR(>F)'][0]

                    for combo in list(combinations(ROI.unique(), 2))
                ]
                model_pairs = [
                    combo
                    for combo in list(combinations(ROI.unique(), 2))
                ]

                fig = plt.figure(figsize=(
                    (10+math.comb(self.ROI_num, 2)/10)/2, (7+math.comb(self.ROI_num, 2)/3)/2))
                plt.rcParams.update(
                    {'font.size': (10+math.comb(self.ROI_num, 2)/10)})
                ax = fig.add_subplot()
                sns.boxplot(x="ROI", y="intensity",
                            data=image_cluster_df_cluster, color='white')
                sns.stripplot(x="ROI", y="intensity", data=image_cluster_df_cluster,
                              color='red', size=4, jitter=True)

                annotator = Annotator(ax, model_pairs, x="ROI",
                                      y="intensity", data=image_cluster_df_cluster)
                formatted_pvalues = [f'{significance(p)}' for p in pvalue]
                annotator.set_custom_annotations(formatted_pvalues)
                annotator.annotate()

                plt.title(f'Mean image of cluster {cluster}', weight='bold')
                plt.xlabel('ROI', weight='bold')
                plt.ylabel('Mean Intensity', weight='bold')
                plt.tight_layout()
                plt.savefig(
                    f"{self.data_dir}/figures/image_cluster/boxplot_cluster{cluster}.jpg", dpi=200)

        image_cluster_all = image_cluster_all.drop(columns=["ROI"])
        if self.jit:
            insitu_2darray = image_cluster_all.to_numpy()
            img_array_1c = np.zeros([insitu_2darray.shape[1]-2, int(max(insitu_2darray[:, -1]))+1,
                                     int(max(insitu_2darray[:, -2]))+1, 1])
            insitu_image = make_image_1c_njit(
                data_2darray=insitu_2darray, img_array_1c=img_array_1c, remap_coord=False)
            del insitu_2darray, img_array_1c
        else:
            insitu_image = make_image_1c(data_2darray=image_cluster_all.to_numpy(),
                                         remap_coord=False, max_normalize=False)

        ROI_info = self.ROI_info[self.ROI_info["replicate"] == replicate]
        ROI_info.reset_index(drop=True, inplace=True)

        if img_plot_method == "plot_img":
            for analyte, im in enumerate(insitu_image[:, :, :, 0]):
                q_max = np.percentile(im, quantile)
                norm = Normalize(vmin=im.min(), vmax=q_max)
                im = norm(im)
                im = np.clip(im, 0, 1)

                plt.style.use('dark_background')
                fig = plt.figure(figsize=np.max(
                    insitu_df[['x', 'y']], axis=0)/10)
                plt.rcParams.update({'font.size': np.max(
                    df_pixel_all_max[['x', 'y']]).max(axis=0)/10})
                gs = GridSpec(1, 2, width_ratios=[np.max(df_pixel_all_max[['x', 'y']]).max(axis=0), np.max(
                    # image and legend grids
                    df_pixel_all_max[['x', 'y']]).max(axis=0)/25], wspace=0.05)
                ax1 = fig.add_subplot(gs[0])
                im_plot = ax1.imshow(im, cmap=color_scheme, vmin=0, vmax=1)
                draw_ROI(ROI_info, show_ROI=show_ROI, show_square=show_square,
                         linewidth=ROI_linewidth, ROI_size_divisor=ROI_size_divisor)
                plt.xticks([])
                plt.yticks([])
                plt.title(f"Mean image of cluster {analyte+1}")
                ax2 = fig.add_subplot(gs[1])
                colorbar = plt.colorbar(im_plot, cax=ax2)
                colorbar.set_ticks([0, 1])
                colorbar.set_ticklabels([0, quantile])
                plt.tight_layout()
                plt.savefig(
                    f"{self.data_dir}/figures/image_cluster/mean_image_cluster{analyte+1}.png")
                plt.show(fig)
            del image_cluster_all
        elif img_plot_method == "plot_ROI":
            show_ROI = True
            show_square = True
            color_scheme = "inferno"
            width_ratios = (
                ROI_info['right']-ROI_info['left']).to_numpy().astype(np.float32)
            width_ratios /= max(width_ratios)
            width_ratios = width_ratios.tolist()
            plt.rcParams.update({'font.size': np.max(
                df_pixel_all_max[['x', 'y']]).max(axis=0)/20})
            ROI_label_size = max(
                ROI_info['right']-ROI_info['left']) / ROI_size_divisor

            for analyte, im in enumerate(insitu_image[:, :, :, 0]):
                q_max = np.percentile(im, quantile)
                norm = Normalize(vmin=im.min(), vmax=q_max)
                im = norm(im)
                im = np.clip(im, 0, 1)

                plt.style.use('dark_background')
                fig = plt.figure()
                gs = GridSpec(1, len(
                    ROI_info) + 1, width_ratios=width_ratios + [0.1], figure=fig, wspace=0.25)

                for ax_index in range(len(ROI_info)):
                    ax = fig.add_subplot(gs[0, ax_index])
                    ROI = pd.DataFrame(ROI_info.loc[ax_index]).T
                    start_col = np.where(ROI.columns == "bottom")[0][0]
                    truncate_col = np.where(ROI_info.columns == "right")[0][0]
                    squares = ROI.iloc[:, start_col:truncate_col+1].to_numpy()

                    im_ROI = im[int(squares[0][0]):(int(squares[0][1])+1),
                                int(squares[0][2]):(int(squares[0][3])+1)]
                    ax.imshow(im_ROI, cmap=color_scheme, vmin=0, vmax=1)
                    plt.sca(ax)
                    for i, square in enumerate(squares):
                        bottom, top, left, right = square
                        top -= bottom
                        bottom -= bottom
                        right -= left
                        left -= left
                        x_coords = [left, left, right, right, left]
                        y_coords = [bottom, top, top, bottom, bottom]
                        if show_square:
                            plt.plot(x_coords, y_coords, 'g-', linewidth=3)
                        if show_ROI:
                            plt.text(right/2, bottom-1, ROI.iloc[0, 0],
                                     horizontalalignment='center', size=ROI_label_size, color='white')

                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.axis('off')

                cax = fig.add_subplot(gs[0, -1])
                fig.suptitle(f"Mean image of cluster \n"
                             f"{analyte+1}", fontsize=ROI_label_size, fontweight='bold')
                scalar_mappable = ScalarMappable(norm=norm, cmap=color_scheme)
                scalar_mappable.set_clim(0, 1)
                colorbar = plt.colorbar(
                    scalar_mappable, cax=cax, orientation='vertical')
                colorbar.set_ticks([0, 1])
                colorbar.set_ticklabels([0, quantile])
                cax.set_position([cax.get_position().x0, cax.get_position().y0 + cax.get_position(
                ).height * 0.25, cax.get_position().width, cax.get_position().height * 0.5])
                cax_pos = cax.get_position()
                line_y = cax_pos.y1 + 0.05
                line_x_start = cax_pos.x0
                line_x_end = cax_pos.x1

                line = Line2D([line_x_start, line_x_end], [line_y, line_y], color='white',
                              transform=fig.transFigure, clip_on=False, linewidth=ROI_linewidth)
                fig.add_artist(line)
                resolution_plt = resolution * \
                    max(round(
                        np.max(ROI_info['right']-ROI_info['left'])*(line_x_end-line_x_start), 0)*self.ROI_num, 1)
                fig.text((line_x_start + line_x_end) / 2, line_y + 0.02,
                         f'{resolution_plt: .0f} $\mu$m', ha='center', va='center', color='white', fontsize=ROI_label_size/2)
                plt.savefig(f"{self.data_dir}/figures/image_cluster/mean_image_cluster\n"
                            f"{analyte+1}.png", bbox_inches='tight')
                plt.show()

        if insitu_tsne:
            print("3D t-SNE not supported on GPU. CPU computing time may be long.")
            truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
            self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].fillna(
                0)
            df_pixel_all_max = self.df_pixel_all.copy()
            df_pixel_all_max = df_pixel_all_max.loc[df_pixel_all_max["ROI"]
                                                    != "placeholder"]
            df_pixel_all_max.reset_index(drop=True, inplace=True)
            df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
                np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=0), axis=1
            )
            df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
                np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=1), axis=0
            )
            df_pixel_all_max = df_pixel_all_max.fillna(0)

            from sklearn.manifold import TSNE
            num_iters = 1000
            my_random_state = 1
            for cluster_index in range(1, k+1):
                truncate_col = np.where(df_pixel_all_max.columns == 'x')[0][0]
                try:
                    tsne_3d = TSNE(n_components=3, n_iter=num_iters,
                                   perplexity=insitu_perplexity, random_state=my_random_state)
                    tsne_embedding_3d = tsne_3d.fit_transform(
                        df_pixel_all_max.iloc[:, :truncate_col].iloc[:, kmeans_labels == cluster_index])
                except ValueError:
                    try:
                        tsne_3d = TSNE(n_components=2, n_iter=num_iters,
                                       perplexity=insitu_perplexity, random_state=my_random_state)
                        tsne_embedding_3d = tsne_3d.fit_transform(
                            df_pixel_all_max.iloc[:, :truncate_col].iloc[:, kmeans_labels == cluster_index])
                        tsne_embedding_3d = np.hstack(
                            (tsne_embedding_3d, np.mean(tsne_embedding_3d, axis=1).reshape((-1, 1))))
                    except ValueError:
                        tsne_3d = TSNE(n_components=1, n_iter=num_iters,
                                       perplexity=insitu_perplexity, random_state=my_random_state)
                        tsne_embedding_3d = tsne_3d.fit_transform(
                            df_pixel_all_max.iloc[:, :truncate_col].iloc[:, kmeans_labels == cluster_index])
                        tsne_embedding_3d = np.column_stack(
                            (tsne_embedding_3d[:, 0], tsne_embedding_3d[:, 0], tsne_embedding_3d[:, 0]))

                rgb_tsne = pd.concat([df_pixel_all_max['replicate'], pd.DataFrame(tsne_embedding_3d),
                                      df_pixel_all_max[['x', 'y']]], axis=1)
                rgb_tsne = rgb_tsne[rgb_tsne['replicate'] == replicate]
                rgb_tsne = rgb_tsne.drop(columns=['replicate'])
                rgb_tsne = rgb_tsne.to_numpy()

                img_array_1c = np.zeros([rgb_tsne.shape[1]-2, int(max(rgb_tsne[:, -1]))+1,
                                         int(max(rgb_tsne[:, -2]))+1, 1])
                insitu_image = make_image_1c_njit(
                    data_2darray=rgb_tsne, img_array_1c=img_array_1c, remap_coord=False)
                insitu_image = np.ma.masked_equal(insitu_image, 0)
                norm = Normalize(vmin=insitu_image.min(),
                                 vmax=insitu_image.max())
                insitu_image = norm(insitu_image)
                insitu_image = np.clip(insitu_image, 0, 1)
                insitu_image = np.transpose(
                    insitu_image[:, :, :, 0], (1, 2, 0))
                width_ratios = (
                    ROI_info['right']-ROI_info['left']).to_numpy().astype(np.float32)
                width_ratios /= max(width_ratios)

                width_ratios = width_ratios.tolist()
                plt.style.use('dark_background')
                fig = plt.figure()
                gs = GridSpec(
                    1, len(ROI_info)+1, width_ratios=width_ratios + [0.1], figure=fig, wspace=0.25)
                ROI_info = self.ROI_info[self.ROI_info["replicate"]
                                         == replicate]
                ROI_info.reset_index(drop=True, inplace=True)

                plt.rcParams.update({'font.size': np.max(
                    df_pixel_all_max[['x', 'y']]).max(axis=0)/20})
                ROI_label_size = max(
                    ROI_info['right']-ROI_info['left']) / ROI_size_divisor
                for ax_index in range(len(ROI_info)):
                    ax = fig.add_subplot(gs[0, ax_index])
                    ROI = pd.DataFrame(ROI_info.loc[ax_index]).T
                    start_col = np.where(ROI.columns == "bottom")[0][0]
                    truncate_col = np.where(ROI_info.columns == "right")[0][0]
                    squares = ROI.iloc[:, start_col:truncate_col+1].to_numpy()

                    im_ROI = insitu_image[int(squares[0][0]):(int(squares[0][1])+1),
                                          int(squares[0][2]):(int(squares[0][3])+1)]
                    ax.imshow(im_ROI, vmin=0, vmax=1)
                    plt.sca(ax)
                    for i, square in enumerate(squares):
                        bottom, top, left, right = square
                        top -= bottom
                        bottom -= bottom
                        right -= left
                        left -= left
                        x_coords = [left, left, right, right, left]
                        y_coords = [bottom, top, top, bottom, bottom]
                        if show_square:
                            plt.plot(x_coords, y_coords, 'g-', linewidth=3)
                        if show_ROI:
                            plt.text(right/2, bottom-1, ROI.iloc[0, 0],
                                     horizontalalignment='center', size=ROI_label_size, color='white')

                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.axis('off')

                plt.savefig(f"{self.data_dir}/figures/image_cluster/insitu_3D_tsne_cluster\n"
                            f"{cluster_index}.png", bbox_inches='tight')
                plt.show()

    def optimize_image_clustering(
        self,
        k_max: int = 10
    ):
        """
        Group ion images by spatial co-localization. Outputs in situ and box plot visualizations of mean ion image for each cluster. Clustering analysis and in situ mapping of clusters are summarized in lower-dimensional t -SNE embedding.

        Parameters
        ----------
        k_max : int, optional
            Maximum number of clusters for cluster validity evaluation
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/image_cluster"):
            os.makedirs(f"{self.data_dir}/figures/image_cluster")

        score_array = np.zeros((k_max-1, 4))

        truncate_col = np.where(self.df_pixel_all.columns == 'x')[0][0]
        self.df_pixel_all.iloc[:, :truncate_col] = self.df_pixel_all.iloc[:, :truncate_col].fillna(
            0)
        df_pixel_all_max = self.df_pixel_all.copy()
        df_pixel_all_max = df_pixel_all_max[df_pixel_all_max["ROI"]
                                            != "placeholder"]
        df_pixel_all_max.reset_index(drop=True, inplace=True)
        df_pixel_all_max.iloc[:, :truncate_col] = df_pixel_all_max.iloc[:, :truncate_col].div(
            np.max(df_pixel_all_max.iloc[:, :truncate_col], axis=0))

        num_iters = 1000
        for k in range(2, k_max+1):
            self.k = k
            if self.gpu == False:
                print("Image clustering on CPU. Computing time may be long.")
                from sklearn.cluster import KMeans
                k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                 n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col].T)
                kmeans_labels = k_means.labels_ + 1

            elif self.gpu == True:

                try:
                    print("Image clustering on GPU.")
                    from cuml import using_output_type
                    from cuml.cluster import KMeans

                    with using_output_type('numpy'):

                        k_means = KMeans(n_clusters=self.k, max_iter=num_iters).fit(
                            df_pixel_all_max.iloc[:, :truncate_col].T)
                        kmeans_labels = k_means.labels_ + 1

                except:
                    print("GPU error. Defaulting to CPU.")
                    print("Image clustering on CPU. Computing time may be long.")
                    from sklearn.cluster import KMeans

                    k_means = KMeans(n_clusters=self.k, init="k-means++", random_state=0,
                                     n_init="auto", max_iter=num_iters).fit(df_pixel_all_max.iloc[:, :truncate_col].T)
                    kmeans_labels = k_means.labels_ + 1

                score_array[k-2, 0] = davies_bouldin_score(
                    df_pixel_all_max.iloc[:, :truncate_col].T, kmeans_labels)
                score_array[k-2, 1] = calinski_harabasz_score(
                    df_pixel_all_max.iloc[:, :truncate_col].T, kmeans_labels)
                score_array[k-2, 2] = silhouette_score(df_pixel_all_max.iloc[:, :truncate_col].T, k_means.fit_predict(
                    df_pixel_all_max.iloc[:, :truncate_col].T))
                score_array[k-2, 3] = k_means.inertia_

        plt.style.use('default')
        categories = ["DB", "CH", "Silhouette", "Elbow"]
        plt.rcParams.update({'font.size': 22})
        fig, ax = plt.subplots(figsize=(10, 7))
        colors = ['blue', 'green', 'red', 'purple', "black"]
        normalized_scores = np.divide(score_array, np.max(score_array, axis=0))
        x_values = np.arange(2, k_max+1)
        for i, (category, color) in enumerate(zip(categories, colors)):
            y_values = normalized_scores[:, i]

            plt.scatter(x_values, y_values, color=color, label=category)
            plt.plot(x_values, y_values, color=color)

        plt.xlabel('Number of clusters', weight='bold')
        plt.ylabel('Max-Normalized Scores', weight='bold')

        plt.legend(title='Score', fontsize="small",
                   title_fontproperties={'weight': 'bold'})
        plt.savefig(
            f"{self.data_dir}/figures/image_cluster/image_clustering_optimization_plot.png", bbox_inches='tight')
        plt.show()

    def make_boxplot(
            self
    ):
        """
        Box plot comparison of mean ion intensity across ROIs with pairwise statistical comparison. Statistical significance thresholds are represented as `*` if 0.05 > $p$-value $geq$ 0.01, `**` if 0.01 > $p$-value $geq$ 0.001, and `***` if $p$-value $\leq$ 0.001.
        """
        if not os.path.exists(f"{self.data_dir}/figures"):
            os.makedirs(f"{self.data_dir}/figures")
        if not os.path.exists(f"{self.data_dir}/figures/boxplot"):
            os.makedirs(f"{self.data_dir}/figures/boxplot")

        self.df_mean_all = self.df_pixel_all.groupby(
            ['ROI', 'replicate', 'ROI_num']).mean().reset_index()
        self.df_mean_all.columns = self.df_mean_all.columns.astype(str)
        self.df_mean_all = self.df_mean_all[self.df_mean_all['ROI']
                                            != 'placeholder']
        if self.df_mean_all['ROI'].shape[0] < 2:
            warnings.warn("Needs at least 2 ROIs.", RuntimeWarning)
            return None

        ROI = self.df_mean_all.iloc[:, 0]
        df_mean_all_long = pd.melt(self.df_mean_all, id_vars=[
                                   'ROI', 'replicate', 'ROI_num', 'x', 'y'], var_name='analyte', value_name='intensity')
        for analyte in df_mean_all_long['analyte'].unique():
            df_mean_all_analyte = df_mean_all_long[df_mean_all_long['analyte'] == analyte]

            try:
                df_resid = ols('intensity ~ C(ROI)', data=df_mean_all_analyte.loc[
                    (df_mean_all_analyte['ROI'] == np.unique(df_mean_all_analyte['ROI'])[0]) |
                    (df_mean_all_analyte['ROI'] ==
                     np.unique(df_mean_all_analyte['ROI'])[1]),
                    ['ROI', 'intensity']]
                ).fit().df_resid
                if df_resid == 0:
                    warnings.warn(
                        "Insufficient sample size for computing p-values.", RuntimeWarning)
                    return None
            except ValueError:
                warnings.warn(
                    "Insufficient sample size for computing p-values.", RuntimeWarning)
                return None

            pvalue = [
                sm.stats.anova_lm(
                    ols('intensity ~ C(ROI)', data=df_mean_all_analyte.loc[
                        (df_mean_all_analyte['ROI'] == combo[0]) |
                        (df_mean_all_analyte['ROI'] == combo[1]),
                        ['ROI', 'intensity']]
                        ).fit(),
                    typ=1
                )['PR(>F)'][0]

                for combo in list(combinations(ROI.unique(), 2))
            ]
            model_pairs = [
                combo
                for combo in list(combinations(ROI.unique(), 2))
            ]
            plt.style.use('default')
            fig = plt.figure(figsize=(
                (10+math.comb(self.ROI_num, 2)/10)/2, (7+math.comb(self.ROI_num, 2)/3)/2))
            ax = fig.add_subplot()
            plt.rcParams.update({'font.size': (7+math.comb(self.ROI_num, 2))})

            sns.boxplot(x="ROI", y="intensity",
                        data=df_mean_all_analyte, color='white')
            sns.stripplot(x="ROI", y="intensity", data=df_mean_all_analyte,
                          color='red', size=4, jitter=True)

            annotator = Annotator(ax, model_pairs, x="ROI",
                                  y="intensity", data=df_mean_all_analyte)
            formatted_pvalues = [f'{significance(p)}' for p in pvalue]
            annotator.set_custom_annotations(formatted_pvalues)
            annotator.annotate()

            plt.title(f'm/z {self.mz.iloc[int(analyte)]}', weight='bold')
            plt.xlabel('ROI', weight='bold')
            plt.ylabel('Mean Intensity', weight='bold')
            plt.tight_layout()
            plt.savefig(
                f"{self.data_dir}/figures/boxplot/mz_{self.mz[int(analyte)]:.3f}.jpg", dpi=200)
            plt.show()



# -*- coding: utf-8 -*-
"""
iMSminer: A Data Processing and Machine Learning Package for Imaging Mass Spectrometry
@author: Yu Tin Lin (yutinlin@stanford.edu)
@author: Haohui Bao (susanab20020911@gmail.com)
@author: Troy R. Scoggins IV (t.scoggins@ufl.edu)
@author: Boone M. Prentice (booneprentice@ufl.chem.edu)
License: Apache-2.0 
"""

import gc
import os
import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import psutil
import scipy
from pyimzml.ImzMLParser import ImzMLParser
from scipy.signal import find_peaks, peak_widths
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures

from .ImzMLParser_chunk import ImzMLParser_chunk
from .utils import (
    Aligner_CPU,
    chunk_prep,
    chunk_prep_inhomogeneous,
    get_chunk_ms_info2,
    get_chunk_ms_info_inhomogeneous,
    get_spectrum,
    integrate_peak,
    str2bool,
)

try:
    import cupy as cp

    from iMSminer.utils import Aligner_GPU
except:
    pass


# ========================DATA PROCESSING========================#


def prompt_for_attributes(prompt_func):
    """
    Decorator for question prompts
    """

    def decorator(method):
        def wrapper(self, *args, **kwargs):
            attributes = method(self, *args, **kwargs)
            for attr_name, prompt in attributes.items():
                # Check if attribute is None or does not exist
                if not hasattr(self, attr_name) or getattr(self, attr_name) is None:
                    setattr(self, attr_name, prompt())
            return method(self, *args, **kwargs)

        return wrapper

    return decorator


def prompt_func(self, attr_name):
    """
    Question prompts for various attributes for class Preprocess()
    """
    prompts = {
        "dist": lambda: int(
            input(
                "\n\n Enter the minimum number of data points between near-isobars: \n\n")
        ),
        "loq": lambda: float(
            input(
                "\n\n What is your limit of quantification? Enter a coefficient k such that LOQ = k * noise: \n\n"
            )
        ),
        "pp_dataset": lambda: input(
            "\n\n Enter dataset name (without '') to perform peak picking on: \n\n"
        ),
        "lwr": lambda: float(input("\n\n Enter a lower m/z bound of noise: \n\n")),
        "upr": lambda: float(input("\n\n Enter an upper m/z bound of noise: \n\n")),
        "noise": lambda: float(
            input(
                "\n\n Specify a noise level. Enter a number noise such that LOQ = k * noise: \n\n"
            )
        ),
        "z_score": lambda: float(
            input("\n\n Enter a z_score bound for noise classification: \n\n")
        ),
        "RP": lambda: float(input("\n\n Specify FWHF: \n\n")),
        "mz_RP": lambda: float(input("\n\n Specify the m/z where FWHF was calculated: \n\n")),
        "rp_factor": lambda: float(input("\n\n Scale number of bins by a number: \n\n")),
    }
    return prompts[attr_name]()


class Preprocess:
    """
    Contains functions to import imzML, generate interactive mean mass spectrum, perform peak picking, mass alignment, and peak integration

    Attributes
    ----------
    directory : str
        Directory that contains all imzML files to preprocess
    data_dir : str
        Directory to save preprocessed data
    gpu : bool
        True if gpu-accelerated libraries are imported successfully
    dist : int, user input
        Minimum number of datapoints for peak separation
    loq : float, user input
        Number of times the noise level (k * noise) to define limit of quantification used in peak picking
    pp_dataset : str, user input
        File name of dataset to perform peak picking on
    lwr : float, user input
        Lower m/z bound of a region in spectrum without signals
    upr : float, user input
        Upper m/z bound of a region in spectrum without signals
    noise : float
        Noise level to guide peak picking
    z_score : float, user input
        Statistical upper threshold for noise computation
    RP : float
        Resolving power [FWHM] used to bin spectra
    mz_RP : float
        m/z at which RP is calculated
    rp_factor : float, user input
        Method `binning`, factor to scale number of bins; affects mass resolution
    resolution_progress : str
        Sets resolution for binning if `yes`
    """

    def __init__(self):
        self.directory = input(
            "\n\n Enter the directory of imzML files to process: \n\n")
        self.data_dir = input(
            "\n\n Enter a directory path for saving preprocessed data: \n\n")

        test_library = ["cupy"]
        for lib in test_library:
            try:
                __import__(lib)
                print(f"{lib.capitalize()} is installed and imported successfully!")
                self.gpu = True
            except ImportError:
                print(
                    f"{lib.capitalize()} is not installed or could not be imported.")
                self.gpu = False
                break
        # Initialize attributes as type None
        self.dist = None
        self.loq = None
        self.pp_dataset = None
        self.lwr = None
        self.upr = None
        self.noise = None
        self.z_score = None
        self.RP = None
        self.mz_RP = None
        self.rp_factor = None
        self.resolution_progress = None

    @prompt_for_attributes(prompt_func)
    def check_attributes(self):
        attributes = {
            "dist": lambda: int(
                input(
                    "\n\n Enter the minimum number of data points required to discern closely-spaced peaks: \n\n"
                )
            ),
            "loq": lambda: float(
                input(
                    "\n\n Specify an intensity threshold for peak picking. Enter a coefficient k such that threshold = k * noise: \n\n"
                )
            ),
            "pp_dataset": lambda: (
                input(
                    "\n\n Enter dataset name (without '') to perform peak picking on: \n\n")
                if self.datasets.shape[0] > 1
                else self.datasets[0]
            ),
            "lwr": lambda: (
                float(input("\n\n Enter a lower m/z bound of noise: \n\n"))
                if self.pp_method == "point"
                else None
            ),
            "upr": lambda: (
                float(input("\n\n Enter an upper m/z bound of noise: \n\n"))
                if self.pp_method == "point"
                else None
            ),
            "noise": lambda: (
                float(
                    input(
                        "\n\n Specify a noise level. Enter a number noise such that LOQ = k * noise: \n\n"
                    )
                )
                if self.pp_method == "specify_noise"
                else None
            ),
            "z_score": lambda: (
                float(
                    input(
                        "\n\n Enter a z_score bound (number of standard deviations) for noise classification: \n\n"
                    )
                )
                if self.pp_method in ["automatic", "binning_even", "binning_regression"]
                else None
            ),
        }
        if self.pp_method in ["binning_even", "binning_regression"]:
            attributes.update(
                {
                    "RP": lambda: float(input("\n\n Specify the resolving power (FWHM): \n\n")),
                    "mz_RP": lambda: float(
                        input(
                            "\n\n Specify the m/z at which resolving power was calculated: \n\n"
                        )
                    ),
                    "rp_factor": lambda: float(
                        input(
                            "\n\n Scale number of m/z bins up or down by a coefficient k (i.e., k * (m/z range of MS analysis) * FWHM / (m/z) = number of bins): \n\n"
                        )
                    ),
                }
            )
        return attributes

    def peak_pick(
        self,
        percent_RAM: float = 5,
        pp_method: str = "point",
        rel_height: float = 0.9,
        peak_alignment: bool = False,
        align_threshold: float = 1,
        align_halfwidth: int = 100,
        grid_iter_num: int = 20,
        align_reduce: bool = False,
        reduce_halfwidth: int = 200,
        plot_aligned_peak: bool = True,
        index_peak_plot: int = 0,
        plot_num_peaks: int = 10,
        baseline_subtract: bool = True,
        baseline_method: str = "noise",
    ):
        """Perform peak picking to locate signals above a defined LOQ (k * noise) by specifying k, calculating noise, and at specified minimum distance between peaks

        Parameters
        ----------
        percent_RAM : int, optional
            Percent available RAM occupied by chunk, by default 5
        pp_method : str, optional
            Method of computing noise, by default "point"
            Method `point` takes specified lower and upper m/z bound of a region in spectrum without signals and compute its standard deviation to define noise level
            Method `specify_noise` takes user-specified noise level
            Method `automatic` computes standard deviation based on spectral data points with a  z-score below a threshold k * z-score, where k is specified by user.
            Method `binning` re-bins mass spectra (userful for compressed data with inhomogeneous shapes), then computes noise using method "automatic"
        rel_height : float
            Peak height cutoff for peak integration, by default 0.9
        peak_alignment : bool
            Performs peak alignment if True, by default False. Peak alignment function refactored from (https://github.com/lukasz-migas/msalign)
        align_threshold : float
            Coefficient to define for peaks for alignment, where peaks above align_threshold*noise are aligned
        align_halfwidth : int
            Half width [data points] to define window for mass alignment around a specified peak
        grid_iter_num : int
            Number of steps to be used in the grid search. Default: 20
        align_reduce : bool
            Reduces size of m/z and intensity arrays used in alignment if True, by default False
        reduce_halfwidth: int
            Half width [data points] to define reduction size of m/z and intensity arrays used in alignment if `align_reduce=True`, by default 200
        plot_aligned_peak : bool
            Plots a specified peak after alignment if True, by default True
        index_peak_plot : int
            Index of peak to plot if `plot_aligned_peak=True`, by default 0
        plot_num_peaks : int
            Number of peaks to plot if `plot_aligned_peak=True`, by deault 10
        baseline_subtract : bool
            Calculates baseline and subtracts all intensities from baseline if `baseline_subtract=True`
        baseline_method : str
            Method of baseline calculation if `baseline_subtract=True`
            Method `regression` defines baseline using polynomial regression of input degree
            Method `noise` defines baseline as input coefficient * noise
        """
        self.pp_method = pp_method
        self.percent_RAM = percent_RAM
        self.align_halfwidth = align_halfwidth
        self.grid_iter_num = grid_iter_num
        self.align_reduce = align_reduce
        self.plot_aligned_peak = plot_aligned_peak
        self.align_threshold = align_threshold
        self.index_peak_plot = index_peak_plot
        self.plot_num_peaks = plot_num_peaks
        self.baseline_subtract = baseline_subtract
        self.baseline_method = baseline_method

        datasets = os.listdir(self.directory)
        datasets = np.asarray(datasets)[
            (np.char.find(datasets, "imzML") != -1)]
        self.datasets = datasets
        self.rel_height = rel_height
        self.check_attributes()

        self.peak_alignment = False
        self.peak_pick_func()
        if peak_alignment:
            self.peak_alignment = peak_alignment
            self.peak_pick_func()

    def run(
        self,
        percent_RAM: float = 5,
        peak_alignment: bool = False,
        integrate_method: str = "peak_width",
        align_halfwidth: int = 100,
        grid_iter_num: int = 20,
        align_reduce: bool = True,
        reduce_halfwidth: int = 200,
        plot_aligned_peak: bool = True,
        index_peak_plot: int = 0,
        plot_num_peaks: int = 10,
    ):
        """imports imzML files and perform peak-picking, mass alignment, and peak integration

        Parameters
        ----------
        percent_RAM : int, optional
            Percent available RAM occupied by chunk, by default 5
        peak_alignment : bool, optional, user input
            Performs mass alignment on peaks detected by peak picking if True, by default False
        align_halfwidth : int, user input
            Half-width of window for alignment, by default 100
        grid_iter_num : int, user input
            Number of steps by grid search, by default 20. Larger values give more accurate quantification results but computation time increases quadratically
        align_reduce : bool, optional
            Reduce the size of intensity matrix passed into alignment if True, by default True
        reduce_halfwidth : int, user input
            Half-width of window around peaks for which intensity matrix is reduced before passing into the mass alignment function if True, by default 200
        plot_aligned_peak : bool, optional
            Render a figure to show peak alignment results if True, by default True
        index_peak_plot : int, user input
            Peak with specified analyte index to visualize if plot_aligned_peak, by default 0
        plot_num_peaks : int, user input
            Number of peaks (spectra) at index_peak_plot to plot if True, by default 10
        """
        self.peak_alignment = peak_alignment
        self.align_halfwidth = align_halfwidth
        self.grid_iter_num = grid_iter_num
        self.align_reduce = align_reduce
        self.plot_aligned_peak = plot_aligned_peak
        self.index_peak_plot = index_peak_plot
        self.plot_num_peaks = plot_num_peaks

        for dataset in self.datasets:
            if not self.inhomogeneous:
                p = ImzMLParser_chunk(f"{self.directory}/{dataset}")
                num_chunks, chunk_size_base, chunk_start, remainder = chunk_prep(
                    p, percent_RAM
                )
            else:
                p = ImzMLParser(f"{self.directory}/{dataset}")
                num_chunks, chunk_size_base, chunk_start, remainder = (
                    chunk_prep_inhomogeneous(p, self.n_bins, percent_RAM)
                )
            num_spectra = len(p.coordinates)
            chunk_start = 0
            previous_chunk_size = 0
            coords_df = pd.DataFrame()

            for i in range(num_chunks):
                start_time = time.time()
                chunk_size_temp = chunk_size_base

                if remainder > i:
                    chunk_size_temp += 1
                print(f"chunk: {i+1} of {num_chunks}; \n"
                      f"chunck_size: {chunk_size_temp} of {num_spectra}")
                if i != 0:
                    chunk_start += previous_chunk_size

                if not self.inhomogeneous:
                    chunk_ms_dict = get_chunk_ms_info2(
                        p, chunk_start, chunk_size_temp)
                else:
                    chunk_ms_dict = get_chunk_ms_info_inhomogeneous(
                        p,
                        chunk_start,
                        chunk_size_temp,
                        self.max_mz,
                        self.min_mz,
                        self.mz_RP,
                        self.RP,
                        self.dist,
                        self.rp_factor,
                    )
                    chunk_ms_dict["mz"] = self.mzs
                if self.baseline_subtract:
                    chunk_ms_dict["I"] -= self.baseline_pred
                    chunk_ms_dict["I"][chunk_ms_dict["I"] < 0] = 0

                chunk_ms_dict["mz"], chunk_ms_dict["I"] = self.peak_alignment_func(
                    mz=chunk_ms_dict["mz"], intensity=chunk_ms_dict["I"]
                )

                # chunk 0 settings
                if i == 0:
                    peaks_df = pd.DataFrame(
                        np.zeros((len(self.p2), 0)), index=chunk_ms_dict["mz"][self.p2]
                    )

                if not self.inhomogeneous:
                    peak_area_df_temp = integrate_peak(
                        chunk_ms_dict,
                        self.p2,
                        self.p2_width,
                        self.dist,
                        chunk_ms_dict["mz"][self.p2],
                        integrate_method=integrate_method,
                    )
                else:
                    peak_area_df_temp = pd.DataFrame(
                        chunk_ms_dict["I"].T[self.p2, :],
                        index=chunk_ms_dict["mz"][self.p2],
                    )

                peak_area_df_temp = peak_area_df_temp.rename(
                    columns=lambda s: str(int(s) + chunk_start)
                )
                coords_df_temp = pd.DataFrame(chunk_ms_dict["coords"])
                coords_df_temp = coords_df_temp.rename(
                    index=lambda s: str(int(s) + chunk_start)
                )
                peaks_df = pd.concat([peaks_df, peak_area_df_temp], axis=1)
                coords_df = pd.concat([coords_df, coords_df_temp], axis=0)

                previous_chunk_size = chunk_size_temp
                print(
                    f"Time used for processing chunk {i}: {time.time()-start_time}")
                del peak_area_df_temp, coords_df_temp
            gc.collect()
            self.peaks_df = peaks_df
            self.coords_df = coords_df

            print("Starting to write to file")
            if os.path.exists(f"{self.data_dir}") == False:
                os.makedirs(f"{self.data_dir}")

            peaks_df.to_csv(
                (f"{self.data_dir}/{dataset[:-6]}.csv"), index=True)
            coords_df.to_csv(
                # 0,1 = x,y
                (f"{self.data_dir}/{dataset[:-6]}_coords.csv"),
                index=True,
            )

            del peaks_df, coords_df

    def peak_pick_func(self):
        """Performs peak picking using method `point`, `specify_noise`, `automatic`, `binning_even` or `binning_regression`, with optional baseline subtraction using method `regression` or `noise`"""

        try:
            self.p = ImzMLParser_chunk(f"{self.directory}/{self.pp_dataset}")
            self.num_chunks, self.chunk_size_base, self.chunk_start, self.remainder = (
                chunk_prep(self.p, self.percent_RAM)
            )
            if self.pp_method == "point":
                p2, p2_width = self.get_p2(method=self.pp_method)
            elif self.pp_method == "specify_noise":
                p2, p2_width = self.get_p2(method=self.pp_method)
            elif self.pp_method == "automatic":
                index_stop = int(self.p.getmz(0).shape[0] * 0.999)
                avg_spectrum = np.zeros(index_stop)
                mzs = self.p.getmz(0)[:index_stop]
                self.mzs = mzs
                noise_array = np.zeros(index_stop)
                for i in range(self.num_chunks):
                    print(f"Chunk {i}")
                    start_time = time.time()
                    chunk_size_temp = self.chunk_size_base
                    if self.remainder > i:
                        chunk_size_temp += 1
                    if i != 0:
                        self.chunk_start += chunk_size_temp

                    intensities_chunk = self.p.get_intensity_chunk(
                        chunk_start=self.chunk_start,
                        chunk_end=(self.chunk_start + chunk_size_temp),
                        index_stop=index_stop,
                    )
                    mzs, intensities_chunk = self.peak_alignment_func(
                        mz=mzs, intensity=intensities_chunk
                    )

                    try:
                        if self.plot_aligned_peak:
                            plt.style.use("classic")
                            for spectrum_index in range(
                                0,
                                intensities_chunk.shape[0],
                                int(intensities_chunk.shape[0] /
                                    self.plot_num_peaks),
                            ):
                                plt.plot(
                                    intensities_chunk[spectrum_index][
                                        self.p2[self.index_peak_plot]
                                        - 20: self.p2[self.index_peak_plot]
                                        + 20
                                    ]
                                )
                                plt.plot(
                                    20,
                                    intensities_chunk[spectrum_index][
                                        self.p2[self.index_peak_plot]
                                    ],
                                    "x",
                                )
                                plt.hlines(
                                    *np.concatenate(
                                        (
                                            [
                                                intensities_chunk[spectrum_index][
                                                    self.p2[self.index_peak_plot]
                                                ]
                                                / 2
                                            ],
                                            np.asarray(self.p2_width)[
                                                2:, self.index_peak_plot
                                            ]
                                            - self.p2[self.index_peak_plot]
                                            + 20,
                                        ),
                                        axis=0,
                                    ),
                                    color="C2",
                                )
                            plt.show()
                    except AttributeError:
                        pass

                    avg_spectrum += np.sum(intensities_chunk, axis=0)
                    noise_chunk = np.std(intensities_chunk, axis=0)
                    noise_array += noise_chunk

                    print(
                        f"Time used for running chunk {i+1} with a size of \n"
                        f"{chunk_size_temp}: {time.time()-start_time}"
                    )
                gc.collect()
                avg_spectrum /= len(self.p.coordinates)
                self.avg_spectrum = avg_spectrum

                noise = (
                    np.mean(
                        noise_array[abs(scipy.stats.zscore(noise_array)) < 3])
                    / self.num_chunks
                )
                self.noise = noise
                self.noise_array = noise_array / self.num_chunks
                self.baseline_subtraction()

                p2 = find_peaks(
                    x=self.avg_spectrum, height=self.loq * noise, distance=self.dist
                )
                p2 = p2[0]
                p2_width = peak_widths(
                    self.avg_spectrum, p2, rel_height=self.rel_height
                )
                p2_width = np.asarray(p2_width)

            self.inhomogeneous = False
            self.p2 = p2
            self.p2_width = p2_width

            get_spectrum(
                output_filepath=f"{self.data_dir}/\n"
                f"{self.pp_dataset[:-6]}_avg_spectrum.html",
                mzs=self.mzs,
                avg_intensity=self.avg_spectrum,
                p2=self.p2,
            )

        except Exception:
            self.baseline_subtract = False
            if self.pp_method not in ["binning_even", "binning_regression"]:
                self.pp_method = input(
                    "\n\n Select binning method: `binning_even` or `binning_regression`. Use `binning_regression`, which requires simple linear regression equation from original data, for more optimal mass resolution. \n\n"
                )
            self.check_attributes()
            self.p = ImzMLParser(f"{self.directory}/{self.pp_dataset}")
            self.num_chunks, self.chunk_size_base, self.chunk_start, self.remainder = (
                chunk_prep(self.p, self.percent_RAM)
            )
            min_mz = 10**10
            max_mz = 0
            for i in range(len(self.p.coordinates)):
                mz = self.p.getspectrum(i)[0]
                if min(mz) < min_mz:
                    min_mz = min(mz)
                if max(mz) > max_mz:
                    max_mz = max(mz)

            self.min_mz = min_mz
            self.max_mz = max_mz
            if self.pp_method == "binning_even":
                resolution_exit = False
                while not resolution_exit:
                    n_bins = (
                        int((max_mz - min_mz) /
                            (self.mz_RP / self.RP) * self.rp_factor)
                        + 1
                    )
                    self.n_bins = n_bins
                    avg_spectrum = np.zeros(n_bins)
                    self.mzs = np.linspace(min_mz, max_mz, num=n_bins)

                    num_chunks, chunk_size_base, chunk_start, remainder = (
                        chunk_prep_inhomogeneous(
                            self.p, n_bins, self.percent_RAM)
                    )
                    previous_chunk_size = 0
                    num_spectra = len(self.p.coordinates)
                    for i in range(num_chunks):
                        chunk_size_temp = chunk_size_base

                        if remainder > i:
                            chunk_size_temp += 1

                        if i != 0:
                            chunk_start += previous_chunk_size
                        print(f"chunk: {i+1} of {num_chunks}; \n"
                              f"chunck_size: {chunk_size_temp} of {num_spectra}")

                        chunk_ms_dict = get_chunk_ms_info_inhomogeneous(
                            self.p,
                            chunk_start,
                            chunk_size_temp,
                            self.max_mz,
                            self.min_mz,
                            self.mz_RP,
                            self.RP,
                            self.dist,
                            self.rp_factor,
                        )
                        self.mzs, chunk_ms_dict["I"] = self.peak_alignment_func(
                            mz=self.mzs, intensity=chunk_ms_dict["I"]
                        )
                        avg_spectrum += np.sum(chunk_ms_dict["I"], axis=0)
                        previous_chunk_size = chunk_size_temp

                    avg_spectrum /= len(self.p.coordinates)

                    self.noise = np.std(
                        avg_spectrum[
                            abs(scipy.stats.zscore(avg_spectrum)) < self.z_score
                        ]
                    )

                    p2 = find_peaks(
                        x=avg_spectrum, height=self.loq * self.noise, distance=self.dist
                    )
                    p2_width = peak_widths(
                        avg_spectrum, p2[0], rel_height=self.rel_height
                    )

                    self.p2 = p2[0]
                    self.p2_width = np.asarray(p2_width)
                    self.avg_spectrum = avg_spectrum
                    self.inhomogeneous = True

                    get_spectrum(
                        output_filepath=f"{self.data_dir}/\n"
                        f"{self.pp_dataset[:-6]}_avg_spectrum.html",
                        mzs=self.mzs,
                        avg_intensity=avg_spectrum,
                        p2=self.p2,
                    )

                    resolution_progress = input(
                        "\n\n Satisfied with resolution? (yes/no) \n\n")
                    if resolution_progress == "yes":
                        resolution_exit = True
                    else:
                        resolution_exit = False
                        self.rp_factor = float(
                            input("\n\n Scale number of bins by a number: \n\n")
                        )
                        self.dist = int(
                            input(
                                "\n\n What is the minimum number of data points between peaks? \n\n"
                            )
                        )

            elif self.pp_method == "binning_regression":
                self.inhomogeneous = True
                try:
                    self.regression_eq
                except AttributeError:
                    ols_exit = False
                    while not ols_exit:
                        regression_eq = input(
                            "\n\n Specify a simple linear regression equation of intercept + coefficient. Enter intercept{one space}coefficient: \n\n"
                        )
                        regression_eq = regression_eq.split(" ")
                        if len(regression_eq) != 2:
                            print(
                                "\n\n Must enter a simple linear regression equation with an intercept and a coefficient for slope. \n\n")
                            ols_exit = False
                        else:
                            ols_exit = True

                    regression_eq = np.asarray(
                        regression_eq).astype(np.float32)
                    self.regression_eq = regression_eq

                resolution_exit = False
                while not resolution_exit:
                    n_bins = (
                        int((max_mz - min_mz) /
                            (self.mz_RP / self.RP) * self.rp_factor)
                        + 1
                    )
                    self.n_bins = n_bins
                    avg_spectrum = np.zeros(n_bins)
                    self.mzs = np.linspace(min_mz, max_mz, num=n_bins)

                    num_chunks, chunk_size_base, chunk_start, remainder = (
                        chunk_prep_inhomogeneous(
                            self.p, n_bins, self.percent_RAM)
                    )
                    previous_chunk_size = 0
                    num_spectra = len(self.p.coordinates)
                    for i in range(num_chunks):
                        chunk_size_temp = chunk_size_base

                        if remainder > i:
                            chunk_size_temp += 1

                        if i != 0:
                            chunk_start += previous_chunk_size
                        print(f"chunk: {i+1} of {num_chunks}; \n"
                              f"chunck_size: {chunk_size_temp} of {num_spectra}")

                        chunk_ms_dict = get_chunk_ms_info_inhomogeneous(
                            self.p,
                            chunk_start,
                            chunk_size_temp,
                            self.max_mz,
                            self.min_mz,
                            self.mz_RP,
                            self.RP,
                            self.dist,
                            self.rp_factor,
                        )

                        avg_spectrum += np.sum(chunk_ms_dict["I"], axis=0)
                        previous_chunk_size = chunk_size_temp

                    avg_spectrum /= len(self.p.coordinates)

                    self.noise = np.std(
                        avg_spectrum[
                            abs(scipy.stats.zscore(avg_spectrum)) < self.z_score
                        ]
                    )

                    p2 = find_peaks(
                        x=avg_spectrum, height=self.loq * self.noise, distance=self.dist
                    )
                    self.p2 = p2[0]
                    p2_half_width = peak_widths(
                        avg_spectrum, p2[0], rel_height=0.5)
                    p2_half_width = np.asarray(p2_half_width)

                    p2_half_width = p2_half_width[2:].T
                    scale_factor = np.ceil(p2_half_width) - p2_half_width

                    self.FWHM = [
                        self.mzs[self.p2[i]]
                        / (
                            self.mzs[np.floor(row[1]).astype(int)]
                            * (1 - scale_factor[i][1])
                            + self.mzs[np.ceil(row[1]).astype(int)]
                            * (scale_factor[i][1])
                            - (
                                self.mzs[np.floor(row[0]).astype(int)]
                                * (1 - scale_factor[i][0])
                                + self.mzs[np.ceil(row[0]).astype(int)]
                                * (scale_factor[i][0])
                            )
                        )
                        for i, row in enumerate(p2_half_width)
                    ]

                    avg_spectrum = np.zeros(n_bins)

                    rp_predicted = self.regression_eq[0] + self.regression_eq[
                        1
                    ] * np.linspace(min_mz, max_mz, n_bins)
                    rp_predicted = rp_predicted / np.sum(rp_predicted)

                    self.mzs = np.interp(
                        np.linspace(0, 1, n_bins),
                        np.cumsum(rp_predicted),
                        np.linspace(min_mz, max_mz, n_bins),
                    )

                    previous_chunk_size = 0
                    chunk_start = 0

                    for i in range(num_chunks):
                        chunk_size_temp = chunk_size_base

                        if remainder > i:
                            chunk_size_temp += 1

                        if i != 0:
                            chunk_start += previous_chunk_size
                        print(f"chunk: {i+1} of {num_chunks}; \n"
                              f"chunck_size: {chunk_size_temp} of {num_spectra}")

                        chunk_ms_dict = get_chunk_ms_info_inhomogeneous(
                            self.p,
                            chunk_start,
                            chunk_size_temp,
                            self.max_mz,
                            self.min_mz,
                            self.mz_RP,
                            self.RP,
                            self.dist,
                            self.rp_factor,
                        )
                        self.mzs, chunk_ms_dict["I"] = self.peak_alignment_func(
                            mz=self.mzs, intensity=chunk_ms_dict["I"]
                        )
                        avg_spectrum += np.sum(chunk_ms_dict["I"], axis=0)
                        previous_chunk_size = chunk_size_temp

                    avg_spectrum /= len(self.p.coordinates)

                    self.noise = np.std(
                        avg_spectrum[
                            abs(scipy.stats.zscore(avg_spectrum)) < self.z_score
                        ]
                    )

                    p2 = find_peaks(
                        x=avg_spectrum, height=self.loq * self.noise, distance=self.dist
                    )
                    p2_width = peak_widths(
                        avg_spectrum, p2[0], rel_height=self.rel_height
                    )

                    self.p2 = p2[0]
                    self.p2_width = np.asarray(p2_width)
                    self.avg_spectrum = avg_spectrum
                    self.inhomogeneous = True

                    get_spectrum(
                        output_filepath=f"{self.data_dir}/\n"
                        f"{self.pp_dataset[:-6]}_avg_spectrum.html",
                        mzs=self.mzs,
                        avg_intensity=avg_spectrum,
                        p2=self.p2,
                    )

                    resolution_progress = input(
                        "\n\n Satisfied with resolution? (yes/no) \n\n")
                    if resolution_progress == "yes":
                        resolution_exit = True
                    else:
                        resolution_exit = False
                        self.rp_factor = float(
                            input("\n\n Scale number of bins by a number: \n\n")
                        )
                        self.dist = int(
                            input(
                                "\n\n What is the minimum number of data points between peaks? \n\n"
                            )
                        )

            else:
                print("Data has inhomogeneous shape. Default to binning.\
                      Mass alignment is recommended for optimal binning. ")
                print(
                    "Unrecognized method. Choose an option from `automatic`, `specify_noise`, or `point` if raw imzML, and `binning_even` or `binning_regression` if reduced imzML."
                )
                return

    def peak_alignment_func(self, mz, intensity):
        """ALigns input intensity array based on peak index or m/z values

        Parameters
        ----------
        mz : np.ndarray
            m/z array used in alignment
        intensity : np.ndarray
            Intensity array before alignment

        Returns
        -------
        mz : np.ndarray
            m/z array used in alignment
        intensity : np.ndarray
            Aligned intesnsity array

        """
        if self.peak_alignment:
            print("aligning . . .")
            if self.gpu:
                gc.collect()
                cp.get_default_memory_pool().free_all_blocks()
                cp.get_default_pinned_memory_pool().free_all_blocks()

                if self.align_reduce:
                    reduced_index = [
                        np.arange(
                            val - self.reduce_halfwidth, val + self.reduce_halfwidth, 1
                        )
                        for val in self.p2
                    ]
                    reduced_index = np.concatenate(reduced_index)
                    reduced_index = reduced_index[reduced_index <
                                                  self.mzs.shape[0] - 1]
                    alinger = Aligner_GPU(
                        x=mz[reduced_index],
                        peaks=mz[self.p2][
                            self.avg_spectrum[self.p2]
                            > self.align_threshold * self.noise
                        ],
                        array=intensity[:, reduced_index],
                        method="gpu_linear",
                        align_by_index=True,
                        only_shift=True,
                        return_shifts=True,
                        width=10,
                        ratio=10,
                        grid_steps=self.grid_iter_num,
                        shift_range=cp.asarray(
                            [-self.align_halfwidth, self.align_halfwidth]
                        ),
                    )

                    alinger.run()
                    aligned_I, shifts = alinger.apply()
                    intensity[:, reduced_index] = aligned_I.get()
                else:
                    alinger = Aligner_GPU(
                        x=mz,
                        peaks=mz[self.p2][
                            self.avg_spectrum[self.p2]
                            > self.align_threshold * self.noise
                        ],
                        array=intensity,
                        method="gpu_linear",
                        align_by_index=True,
                        only_shift=True,
                        return_shifts=True,
                        width=10,
                        ratio=10,
                        grid_steps=self.grid_iter_num,
                        shift_range=cp.asarray(
                            [-self.align_halfwidth, self.align_halfwidth]
                        ),
                    )

                    alinger.run()
                    aligned_I, shifts = alinger.apply()
                    intensity = aligned_I.get()
            else:
                if self.align_reduce:
                    reduced_index = [
                        np.arange(
                            val - self.reduce_halfwidth, val + self.reduce_halfwidth, 1
                        )
                        for val in self.p2
                    ]
                    reduced_index = np.concatenate(reduced_index)
                    reduced_index = reduced_index[reduced_index <
                                                  self.mzs.shape[0] - 1]
                    alinger = Aligner_CPU(
                        x=mz[reduced_index],
                        peaks=mz[self.p2][
                            self.avg_spectrum[self.p2]
                            > self.align_threshold * self.noise
                        ],
                        array=intensity[:, reduced_index],
                        method="linear",
                        align_by_index=True,
                        only_shift=True,
                        return_shifts=True,
                        width=10,
                        ratio=10,
                        grid_steps=self.grid_iter_num,
                        shift_range=cp.asarray(
                            [-self.align_halfwidth, self.align_halfwidth]
                        ),
                    )

                    alinger.run()
                    aligned_I, shifts = alinger.apply()
                    intensity[:, reduced_index] = aligned_I
                else:
                    gc.collect()

                    alinger = Aligner_CPU(
                        x=mz,
                        peaks=mz[self.p2][
                            self.avg_spectrum[self.p2]
                            > self.align_threshold * self.noise
                        ],
                        array=intensity,
                        method="linear",
                        align_by_index=True,
                        only_shift=False,
                        return_shifts=True,
                        width=10,
                        ratio=10,
                        grid_steps=self.grid_iter_num,
                        shift_range=np.asarray(
                            [-self.align_halfwidth, self.align_halfwidth]
                        ),
                    )

                    alinger.run()
                    aligned_I, shifts = alinger.apply()
                    intensity = aligned_I

            del aligned_I

        return mz, intensity

    def get_p2(self):
        """Calculates noise level on average spectrum, performs peak picking, and calculates peak widths for peak integration

        Returns
        -------
        p2 : np.1darray
            m/z bin indices corresponding to peak-picked maxima
        p2_width : np.1darray
            Peak regions computed by np.peak_widths
        """
        index_stop = int(self.p.getmz(0).shape[0] * 0.999)
        avg_spectrum = np.zeros(index_stop)
        mzs = self.p.getmz(0)[:index_stop]
        noise = 0
        for i in range(self.num_chunks):
            print(f"Chunk {i}")
            start_time = time.time()
            chunk_size_temp = self.chunk_size_base
            if self.remainder > i:
                chunk_size_temp += 1

            intensities_chunk = self.p.get_intensity_chunk(
                chunk_start=self.chunk_start,
                chunk_end=(self.chunk_start + chunk_size_temp),
                index_stop=index_stop,
            )

            mzs, intensities_chunk = self.peak_alignment_func(
                mz=mzs, intensity=intensities_chunk
            )

            avg_spectrum += np.sum(intensities_chunk, axis=0)

            if self.pp_method == "point":
                noise_bin = np.std(
                    np.asarray(intensities_chunk)[
                        :, np.logical_and(mzs >= self.lwr, mzs <= self.upr)
                    ],
                    axis=0,
                )
                if np.sum(noise_bin) == 0:
                    pass
                else:
                    noise += np.mean(noise_bin[noise_bin != 0])
                    if i == (self.num_chunks - 1):
                        self.noise /= self.num_chunks

                self.chunk_start += chunk_size_temp

            print(
                f"Time used for running chunk {i} with a size of \n"
                f"{chunk_size_temp}: {time.time()-start_time}"
            )
            gc.collect()

        avg_spectrum /= len(self.p.coordinates)
        self.avg_spectrum = avg_spectrum
        self.mzs = mzs
        p2, info = find_peaks(
            x=avg_spectrum, height=self.loq * self.noise, distance=self.dist
        )
        p2_width = peak_widths(avg_spectrum, p2, rel_height=self.rel_height)
        p2_width = np.asarray(p2_width)

        return p2, p2_width

    def baseline_subtraction(self):
        try:
            self.baseline_pred
        except AttributeError:
            if self.baseline_subtract:
                if self.baseline_method == "regression":
                    baseline_exit = False
                    while not baseline_exit:
                        baseline_index = abs(
                            scipy.stats.zscore(self.noise_array)) < 2
                        degree = int(
                            input(
                                "\n\n Enter degree of regression for baseline computation: \n\n"
                            )
                        )
                        poly = PolynomialFeatures(
                            degree=degree, include_bias=False)
                        poly_features = poly.fit_transform(
                            self.mzs[baseline_index].reshape(-1, 1)
                        )
                        poly_reg_model = LinearRegression()
                        poly_reg_model.fit(
                            poly_features, self.noise_array[baseline_index]
                        )
                        R2 = poly_reg_model.score(
                            poly_features, self.noise_array[baseline_index]
                        )

                        poly_features = poly.fit_transform(
                            self.mzs.reshape(-1, 1))
                        baseline_pred = poly_reg_model.predict(poly_features)

                        fig, ax = plt.subplots(figsize=(30, 10))
                        ax.scatter(self.mzs, baseline_pred, color="red")
                        plt.plot(
                            self.mzs[baseline_index],
                            self.noise_array[baseline_index],
                            alpha=0.3,
                            color="black",
                        )
                        ax.set_title(f"Linear model of degree {degree}")
                        ax.set_xlabel("m/z")
                        ax.set_ylabel("Intensity")
                        ax.text(
                            0.5,
                            0.95,
                            f"$R^2 = {R2:.3f}$",
                            fontsize=12,
                            horizontalalignment="center",
                            verticalalignment="top",
                            transform=plt.gca().transAxes,
                        )
                        plt.show()

                        baseline_exit = input(
                            "\n\n Satisfied with baseline subtraction? Enter (yes / no): \n\n"
                        )
                        baseline_exit = str2bool(baseline_exit)
                    self.baseline_pred = baseline_pred
                elif self.baseline_method == "noise":
                    baseline_factor = float(
                        input(
                            "\n\n Enter a coefficient k for baseline, such that baseline = k * noise: \n\n"
                        )
                    )
                    self.baseline_pred = self.noise * baseline_factor
                self.avg_spectrum -= self.baseline_pred
                self.avg_spectrum[self.avg_spectrum < 0] = 0



import pathlib

import setuptools
from setuptools import find_packages, setup

setuptools.setup(
    name="iMSminer",
    version="1.0.5",
    description="iMSminer provides user-friendly, partially GPU- or compiler-accelerated multi-condition, multi-ROI, and multi-dataset preprocessing and mining of larger-than-memory imaging mass spectrometry datasets in Python.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/Prentice-lab-UF/iMSminer",
    author="Yu Tin Lin",
    author_email="yutinlin@stanford.edu",
    license="Apache Software License",
    install_requires=[
        "bokeh>=3.2.1",
        "opencv-python>=4.5.0",
        "matplotlib>=3.5.0",
        "msalign>=0.2.0",
        "networkx>=3.0.0",
        "numba>=0.50.1",
        "numpy>=1.20.4,<1.26.1",
        "pandas<=1.5.3",
        "psutil>=5.8.0",
        "pyimzml>=1.5.4",
        "scikit-learn>=1.2.2",
        "scipy>=1.11.4",
        "seaborn>=0.11.2",
        "statsmodels>=0.14.0",
        "statannotations>=0.6.0",

    ],
    classifier=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3"
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
)
